# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
The types of messages envelopes supported by Thrift.

.. py:data:: CALL
    :annotation: = 1

    An outgoing call to a specific method.

    ``Message.body`` is a struct containing the request arguments.

.. py:data:: REPLY
    :annotation: = 2

    A response to a ``CALL`` message.

    ``Message.body`` is a union of the return value (with field ID 0) and the
    application exceptions that the message can raise.

.. py:data:: EXCEPTION
    :annotation: = 3

    An unexpected exception in response to a ``CALL`` message.

    ``Message.body`` is a struct representing the exception.

    Note that exceptions that are defined in the IDL are returned as part of
    the ``REPLY`` message. This message type is used for unexpected exceptions
    that were not defined in the IDL but which the server and client have
    agreed upon as standard exceptions that they both recognize.

.. py:data:: ONEWAY
    :annotation: = 4

    An outgoing request to a specific ``oneway`` method.

    ``Message.body`` is the same as a ``CALL`` message but no ``REPLY`` or
    ``EXCEPTION`` is expected in response.

.. versionadded:: 1.0
"""

from __future__ import absolute_import, unicode_literals, print_function

MESSAGE_TYPES = (CALL, REPLY, EXCEPTION, ONEWAY)


cpdef str name_of(int value):
    """Returns the name of the message type with the given value.

    Returns None if no such message type exists.
    """
    if value == CALL:
        return str('CALL')
    elif value == REPLY:
        return str('REPLY')
    elif value == EXCEPTION:
        return str('EXCEPTION')
    elif value == ONEWAY:
        return str('ONEWAY')
    else:
        return None



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from libc.stdint cimport (
    int8_t,
    int16_t,
    int32_t,
    int64_t,
)

from thriftrw.wire cimport ttype
from thriftrw.wire.message cimport Message

from thriftrw._buffer cimport ReadBuffer
from thriftrw._buffer cimport WriteBuffer
from thriftrw.wire.value cimport (
    Value,
    BoolValue,
    ByteValue,
    DoubleValue,
    I16Value,
    I32Value,
    I64Value,
    BinaryValue,
    FieldValue,
    StructValue,
    MapValue,
    MapItem,
    SetValue,
    ListValue,
    ValueVisitor,
)

from .core cimport (
    Protocol,
    ProtocolWriter,
    ProtocolReader,
    FieldHeader,
    MapHeader,
    SetHeader,
    ListHeader,
    MessageHeader,
)
from ..errors import EndOfInputError
from ..errors import ThriftProtocolError
from ._endian cimport (
    htobe16,
    htobe32,
    htobe64,
    be16toh,
    be32toh,
    be64toh,
)


cdef STRUCT_END_HEADER = FieldHeader(-1, -1)

cdef int8_t STRUCT_END = 0

cdef int32_t VERSION = 1

# Under strict mode, the version number is in most significant 16-bits of the
# 32-bit integer, and the most significant bit is set. This mask gets &-ed
# with the integer to get just the version number (it still needs to be
# shifted 16-bits to the right).
cdef int32_t VERSION_MASK = 0x7fff0000

# The least significant 8-bits of the 32-bit integer contain the type.
cdef int32_t TYPE_MASK = 0x000000ff


cdef class BinaryProtocolReader(ProtocolReader):
    """Parser for the binary protocol."""

    def __cinit__(self, ReadBuffer reader):
        """Initialize the reader.

        :param reader:
            File-like object with a ``read(num)`` method which returns
            *exactly* the requested number of bytes, or all remaining bytes if
            ``num`` is negative.
        """
        self.reader = reader

    cdef void skip(self, int typ) except *:
        if typ == ttype.BOOL:
            self.reader.skip(1)
        elif typ == ttype.BYTE:
            self.reader.skip(1)
        elif typ == ttype.DOUBLE:
            self.reader.skip(8)
        elif typ == ttype.I16:
            self.reader.skip(2)
        elif typ == ttype.I32:
            self.reader.skip(4)
        elif typ == ttype.I64:
            self.reader.skip(8)
        elif typ == ttype.BINARY:
            self.skip_binary()
        elif typ == ttype.STRUCT:
            self.skip_struct()
        elif typ == ttype.MAP:
            self.skip_map()
        elif typ == ttype.SET:
            self.skip_set()
        elif typ == ttype.LIST:
            self.skip_list()

    cdef void skip_struct(self) except *:
        header = self.read_field_begin()
        while header.type != -1:
            self.skip(header.type)
            header = self.read_field_begin()

    cdef void skip_map(self) except *:
        # TODO: This could be optimized to precompute the expected
        # size for fixed width types
        cdef MapHeader header = self.read_map_begin()
        for _ in range(header.size):
            self.skip(header.ktype)
            self.skip(header.vtype)

    cdef void skip_list(self) except *:
        # TODO: This could be optimized to precompute the expected
        # size for fixed width types
        cdef ListHeader header = self.read_list_begin()
        for _ in range(header.size):
            self.skip(header.type)

    cdef void skip_set(self) except *:
        # TODO: This could be optimized to precompute the expected
        # size for fixed width types
        cdef SetHeader header = self.read_set_begin()
        for _ in range(header.size):
            self.skip(header.type)

    cdef void _read(self, char* data, int count) except *:
        self.reader.read(data, count)

    cdef int8_t _byte(self) except *:
        cdef char c
        self._read(&c, 1)
        return <int8_t>c

    cdef int16_t _i16(self) except *:
        cdef char data[2]
        self._read(data, 2)
        return be16toh((<int16_t*>data)[0])

    cdef int32_t _i32(self) except *:
        cdef char data[4]
        self._read(data, 4)
        return be32toh((<int32_t*>data)[0])

    cdef int64_t _i64(self) except *:
        cdef char data[8]
        self._read(data, 8)
        return be64toh((<int64_t*>data)[0])

    cdef double _double(self) except *:
        cdef int64_t value = self._i64()
        return (<double*>(&value))[0]

    cdef bint read_bool(self) except *:
        """Reads a boolean."""
        return self._byte() == 1

    cdef int8_t read_byte(self) except *:
        """Reads a byte."""
        return self._byte()

    cdef double read_double(self) except *:
        """Reads a double."""
        return self._double()

    cdef int16_t read_i16(self) except *:
        """Reads a 16-bit integer."""
        return self._i16()

    cdef int32_t read_i32(self) except *:
        """Reads a 32-bit integer."""
        return self._i32()

    cdef int64_t read_i64(self) except *:
        """Reads a 64-bit integer."""
        return self._i64()

    cdef bytes read_binary(self):
        """Reads a binary blob."""
        cdef int32_t length = self._i32()
        return self.reader.take(length)

    cdef void skip_binary(self) except *:
        cdef int32_t length = self._i32()
        self.reader.skip(length)

    cdef FieldHeader read_field_begin(self) except *:
        cdef int8_t field_type = self._byte()
        if field_type == STRUCT_END:
            return STRUCT_END_HEADER

        cdef int16_t field_id = self._i16()

        return FieldHeader(field_type, field_id)

    cdef MapHeader read_map_begin(self) except *:
        cdef int8_t key_ttype = self._byte()
        cdef int8_t value_ttype = self._byte()
        cdef int32_t length = self._i32()

        return MapHeader(key_ttype, value_ttype, length)

    cdef SetHeader read_set_begin(self) except *:
        cdef int8_t value_ttype = self._byte()
        cdef int32_t length = self._i32()

        return SetHeader(value_ttype, length)

    cdef ListHeader read_list_begin(self) except *:
        cdef int8_t value_ttype = self._byte()
        cdef int32_t length = self._i32()

        return ListHeader(value_ttype, length)

    cdef MessageHeader read_message_begin(self):
        cdef int8_t typ
        cdef int16_t version
        cdef int32_t size
        cdef int32_t seqid
        cdef bytes name

        size = self._i32()
        # TODO with cython, some of the Python-specific hacks around bit
        # twiddling can be skipped.
        if size < 0:
            # strict version:
            #
            #     versionAndType:4 name~4 seqid:4 payload
            version = (size & VERSION_MASK) >> 16
            if version != VERSION:
                raise ThriftProtocolError(
                    'Unsupported version "%r"' % version
                )
            typ = size & TYPE_MASK
            size = self._i32()
            name = self.reader.take(size)
        else:
            # non-strict version:
            #
            #     name:4 type:1 seqid:4 payload
            name = self.reader.take(size)
            typ = self._byte()

        seqid = self._i32()
        return MessageHeader(name, typ, seqid)


cdef class BinaryProtocolWriter(ProtocolWriter):
    """Serializes values using the Thrift Binary protocol."""

    def __cinit__(BinaryProtocolWriter self, WriteBuffer writer):
        """Initialize the writer.

        :param writer:
            WriteBuffer to which requests will be written.
        """
        self.writer = writer

    cdef void _write(BinaryProtocolWriter self, char* data, int length):
        self.writer.write(data, length)

    cdef void write_bool(self, bint value) except *:  # bool:1
        self.write_byte(value)

    cdef void write_byte(BinaryProtocolWriter self, int8_t value) except *:
        # byte:1
        cdef char c = <char>value
        self._write(&c, 1)

    cdef void write_double(self, double value) except *:  # double:8
        # If confused:
        #
        #   <typ*>(&value)[0]
        #
        # Is just "interpret the in-memory representation of value as typ"
        self.write_i64((<int64_t*>(&value))[0])

    cdef void write_i16(BinaryProtocolWriter self, int16_t value) except *:
        # i16:2
        value = htobe16(value)
        self._write(<char*>(&value), 2)

    cdef void write_i32(BinaryProtocolWriter self, int32_t value) except *:
        # i32:4
        value = htobe32(value)
        self._write(<char*>(&value), 4)

    cdef void write_i64(BinaryProtocolWriter self, int64_t value) except *:
        # i64:8
        value = htobe64(value)
        self._write(<char*>(&value), 8)

    cdef void write_binary(BinaryProtocolWriter self, char* value, int32_t length) except *:
        # len:4 str:len
        self.write_i32(length)
        self._write(value, length)

    cdef void write_field_begin(BinaryProtocolWriter self,
                                FieldHeader header) except *:
        # type:1 id:2
        self.write_byte(header.type)
        self.write_i16(header.id)

    cdef void write_struct_end(BinaryProtocolWriter self) except *:
        self.write_byte(STRUCT_END)

    cdef void write_map_begin(BinaryProtocolWriter self,
                              MapHeader header) except *:
        # key_type:1 value_type:1 count:4
        self.write_byte(header.ktype)
        self.write_byte(header.vtype)
        self.write_i32(header.size)

    cdef void write_set_begin(BinaryProtocolWriter self,
                              SetHeader header) except *:
        # value_type:1 count:4
        self.write_byte(header.type)
        self.write_i32(header.size)

    cdef void write_list_begin(BinaryProtocolWriter self,
                               ListHeader header) except *:
        # value_type:1 count:4
        self.write_byte(header.type)
        self.write_i32(header.size)

    cdef void write_message_begin(BinaryProtocolWriter self,
                                  MessageHeader message) except *:
        self.write_binary(message.name, len(message.name))
        self.write_byte(message.type)
        self.write_i32(message.seqid)


cdef class BinaryProtocol(Protocol):
    """Implements the Thrift binary protocol."""

    cpdef ProtocolReader reader(self, ReadBuffer buff):
        return BinaryProtocolReader(buff)

    cpdef ProtocolWriter writer(self, WriteBuffer buff):
        return BinaryProtocolWriter(buff)

    cpdef Message deserialize_message(self, bytes s):
        cdef ReadBuffer buff = ReadBuffer(s)
        return _OldBinaryProtocolReader(buff).read_message()

    cpdef Value deserialize_value(self, int typ, bytes s):
        cdef ReadBuffer buff = ReadBuffer(s)
        return _OldBinaryProtocolReader(buff).read(typ)


__all__ = ['BinaryProtocol']


cdef class _OldBinaryProtocolReader(object):
    """Old version of BinaryProtocolReader that returned Value representations.

    Deserialization was refactored in 1.4 to avoid this step as an optimization.
    This class exists to provide support for backwards compatibility but will
    be removed in a later version.

    .. deprecated:: 1.4
    """

    def __cinit__(self, ReadBuffer reader):
        """Initialize the reader.
        :param reader:
            File-like object with a ``read(num)`` method which returns
            *exactly* the requested number of bytes, or all remaining bytes if
            ``num`` is negative.
        """
        self.reader = reader

    cdef object _reader(self, int8_t typ):
        if typ == ttype.BOOL:
            return self.read_bool
        elif typ == ttype.BYTE:
            return self.read_byte
        elif typ == ttype.DOUBLE:
            return self.read_double
        elif typ == ttype.I16:
            return self.read_i16
        elif typ == ttype.I32:
            return self.read_i32
        elif typ == ttype.I64:
            return self.read_i64
        elif typ == ttype.BINARY:
            return self.read_binary
        elif typ == ttype.STRUCT:
            return self.read_struct
        elif typ == ttype.MAP:
            return self.read_map
        elif typ == ttype.SET:
            return self.read_set
        elif typ == ttype.LIST:
            return self.read_list
        else:
            raise ThriftProtocolError('Unknown TType "%r"' % typ)

    cpdef object read(self, int8_t typ):
        if typ == ttype.BOOL:
            return self.read_bool()
        elif typ == ttype.BYTE:
            return self.read_byte()
        elif typ == ttype.DOUBLE:
            return self.read_double()
        elif typ == ttype.I16:
            return self.read_i16()
        elif typ == ttype.I32:
            return self.read_i32()
        elif typ == ttype.I64:
            return self.read_i64()
        elif typ == ttype.BINARY:
            return self.read_binary()
        elif typ == ttype.STRUCT:
            return self.read_struct()
        elif typ == ttype.MAP:
            return self.read_map()
        elif typ == ttype.SET:
            return self.read_set()
        elif typ == ttype.LIST:
            return self.read_list()
        else:
            raise ThriftProtocolError('Unknown TType "%r"' % typ)

    cdef void _read(self, char* data, int count) except *:
        self.reader.read(data, count)

    cdef int8_t _byte(self) except *:
        cdef char c
        self._read(&c, 1)
        return <int8_t>c

    cdef int16_t _i16(self) except *:
        cdef char data[2]
        self._read(data, 2)
        return be16toh((<int16_t*>data)[0])

    cdef int32_t _i32(self) except *:
        cdef char data[4]
        self._read(data, 4)
        return be32toh((<int32_t*>data)[0])

    cdef int64_t _i64(self) except *:
        cdef char data[8]
        self._read(data, 8)
        return be64toh((<int64_t*>data)[0])

    cdef double _double(self) except *:
        cdef int64_t value = self._i64()
        return (<double*>(&value))[0]

    cdef Message read_message(self):
        cdef int8_t typ
        cdef int16_t version
        cdef int32_t size
        cdef bytes name

        size = self._i32()
        # TODO with cython, some of the Python-specific hacks around bit
        # twiddling can be skipped.
        if size < 0:
            # strict version:
            #
            #     versionAndType:4 name~4 seqid:4 payload
            version = (size & VERSION_MASK) >> 16
            if version != VERSION:
                raise ThriftProtocolError(
                    'Unsupported version "%r"' % version
                )
            typ = size & TYPE_MASK
            size = self._i32()
            name = self.reader.take(size)
        else:
            # non-strict version:
            #
            #     name:4 type:1 seqid:4 payload
            name = self.reader.take(size)
            typ = self._byte()

        seqid = self._i32()
        body = self.read(ttype.STRUCT)

        return Message(name=name, seqid=seqid, body=body, message_type=typ)

    cdef BoolValue read_bool(self):
        """Reads a boolean."""
        return BoolValue(self._byte() == 1)

    cdef ByteValue read_byte(self):
        """Reads a byte."""
        return ByteValue(self._byte())

    cdef DoubleValue read_double(self):
        """Reads a double."""
        return DoubleValue(self._double())

    cdef I16Value read_i16(self):
        """Reads a 16-bit integer."""
        return I16Value(self._i16())

    cdef I32Value read_i32(self):
        """Reads a 32-bit integer."""
        return I32Value(self._i32())

    cdef I64Value read_i64(self):
        """Reads a 64-bit integer."""
        return I64Value(self._i64())

    cdef BinaryValue read_binary(self):
        """Reads a binary blob."""
        cdef int32_t length = self._i32()
        return BinaryValue(self.reader.take(length))

    cdef StructValue read_struct(self):
        """Reads an arbitrary Thrift struct."""
        fields = []

        cdef int8_t field_type = self._byte()
        cdef int16_t field_id

        while field_type != STRUCT_END:
            field_id = self._i16()
            field_value = self.read(field_type)
            fields.append(
                FieldValue(
                    id=field_id,
                    ttype=field_type,
                    value=field_value,
                )
            )

            field_type = self._byte()
        return StructValue(fields)

    cdef MapValue read_map(self):
        """Reads a map."""
        cdef int8_t key_ttype = self._byte()
        cdef int8_t value_ttype = self._byte()
        cdef int32_t length = self._i32()

        kreader = self._reader(key_ttype)
        vreader = self._reader(value_ttype)

        pairs = []
        for i in range(length):
            k = kreader(self)
            v = vreader(self)
            pairs.append(MapItem(k, v))

        return MapValue(
            key_ttype=key_ttype,
            value_ttype=value_ttype,
            pairs=pairs,
        )

    cdef SetValue read_set(self):
        """Reads a set."""
        cdef int8_t value_ttype = self._byte()
        cdef int32_t length = self._i32()

        vreader = self._reader(value_ttype)

        values = []

        for i in range(length):
            values.append(vreader(self))

        return SetValue(
            value_ttype=value_ttype,
            values=values
        )

    cdef ListValue read_list(self):
        """Reads a list."""
        cdef int8_t value_ttype = self._byte()
        cdef int32_t length = self._i32()

        vreader = self._reader(value_ttype)

        values = []

        for i in range(length):
            values.append(vreader(self))

        return ListValue(
            value_ttype=value_ttype,
            values=values
        )



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import os.path

from .scope import Scope
from .generate import Generator
from .link import TypeSpecLinker
from .link import ConstSpecLinker
from .link import ServiceSpecLinker
from ..errors import ThriftCompilerError

from thriftrw.idl import Parser
from thriftrw._runtime import Serializer, Deserializer


__all__ = ['Compiler']

LINKERS = [ConstSpecLinker, TypeSpecLinker, ServiceSpecLinker]


class ModuleSpec(object):
    """Specification for a single module."""

    __slots__ = (
        'name',
        'path',
        'scope',
        'surface',
        'includes',
        'protocol',
        'linked',
        'thrift_source',
    )

    def __init__(self, name, protocol, path=None, thrift_source=None):
        """
        :param name:
            Name of the module.
        :param path:
            Path to the Thrift file from which this module was compiled. This
            may be omitted if the module was compiled from an inline string
            (using the ``loads()`` API, for example).
        :param thrift_source:
            If non-None, this is the source code of the Thrift file from which
            this module was generated.
        """

        self.name = name
        self.path = path
        self.protocol = protocol
        self.linked = False
        self.scope = Scope(name, path)
        self.surface = self.scope.module
        self.thrift_source = thrift_source

        # Mapping of names of inculded modules to their corresponding specs.
        self.includes = {}

        # TODO Scope can probably be eventually folded into this class.

    @property
    def can_include(self):
        """Whether this module is allowed to include other modules.

        This is allowed only if the module was compiled from a file since
        include paths are relative to the file in which they are mentioned.
        """
        return self.path is not None

    def add_include(self, name, module_spec):
        """Adds a module as an included module.

        :param name:
            Name under which the included module should be exposed in the
            current module.
        :param module_spec:
            ModuleSpec of the included module.
        """
        assert name, 'name is required'
        assert self.can_include

        if name in self.includes:
            raise ThriftCompilerError(
                'Cannot include module "%s" as "%s" in "%s". '
                'The name is already taken.'
                % (module_spec.name, name, self.path)
            )

        self.includes[name] = module_spec
        self.scope.add_include(name, module_spec.scope, module_spec.surface)

    def link(self):
        """Link all the types in this module and all included modules."""
        if self.linked:
            return self

        self.linked = True

        included_modules = []

        # Link includes
        for include in self.includes.values():
            included_modules.append(include.link().surface)

        self.scope.add_surface('__includes__', tuple(included_modules))
        self.scope.add_surface('__thrift_source__', self.thrift_source)

        # Link self
        for linker in LINKERS:
            linker(self.scope).link()

        self.scope.add_surface('loads', Deserializer(self.protocol))
        self.scope.add_surface('dumps', Serializer(self.protocol))

        return self


class Compiler(object):
    """Compiles IDLs into Python modules."""

    __slots__ = (
        'protocol', 'strict', 'parser', 'include_as', '_module_specs'
    )

    def __init__(self, protocol, strict=None, include_as=None):
        """Initialize the compiler.

        :param thriftrw.protocol.Protocol protocol:
           The protocol ot use to serialize and deserialize values.
        """
        if strict is None:
            strict = True
        if include_as is None:
            include_as = False
        self.protocol = protocol
        self.strict = strict
        self.include_as = include_as

        self.parser = Parser()

        # Mapping from absolute file path to ModuleSpec for all modules.
        self._module_specs = {}

    def compile(self, name, contents, path=None):
        """Compile the given Thrift document into a Python module.

        The generated module contains,

        .. py:attribute:: __services__

            A collection of generated classes for all services defined in the
            thrift file.

            .. versionchanged:: 1.0

                Renamed from ``services`` to ``__services__``.

        .. py:attribute:: __types__

            A collection of generated types for all types defined in the
            thrift file.

            .. versionchanged:: 1.0

                Renamed from ``types`` to ``__types__``.

        .. py:attribute:: __includes__

            A collection of modules included by this module.

            .. versionadded:: 1.0

        .. py:attribute:: __constants__

            A mapping of constant name to value for all constants defined in
            the thrift file.

            .. versionchanged:: 1.0

                Renamed from ``constants`` to ``__constants__``.

        .. py:attribute:: __thrift_source__

            Contents of the .thrift file from which this module was compiled.

            .. versionadded:: 1.1

        .. py:function:: dumps(obj)

            Serializes the given object using the protocol the compiler was
            instantiated with.

        .. py:function:: loads(cls, payload)

            Deserializes an object of type ``cls`` from ``payload`` using the
            protocol the compiler was instantiated with.

        .. py:function:: dumps.message(obj, seqid=0)

            Serializes the given request or response into a
            :py:class:`~thriftrw.wire.Message` using the protocol that the
            compiler was instantiated with.

            See :ref:`calling-apache-thrift`.

            .. versionadded:: 1.0

        .. py:function:: loads.message(service, payload)

            Deserializes a :py:class:`~thriftrw.wire.Message`  from
            ``payload`` using the protocol the compiler was instantiated with.
            A request or response of a method defined in the given service is
            parsed in the message body.

            See :ref:`calling-apache-thrift`.

            .. versionadded:: 1.0

        And one class each for every struct, union, exception, enum, and
        service defined in the IDL.

        Service classes have references to
        :py:class:`thriftrw.spec.ServiceFunction` objects for each method
        defined in the service.

        :param str name:
            Name of the Thrift document. This will be the name of the
            generated module.
        :param str contents:
            Thrift document to compile
        :param str path:
            Path to the Thrift file being compiled. If not specified, imports
            from within the Thrift file will be disallowed.
        :returns:
            ModuleSpec of the generated module.
        """
        assert name

        if path:
            path = os.path.abspath(path)
            if path in self._module_specs:
                return self._module_specs[path]

        module_spec = ModuleSpec(name, self.protocol, path, contents)
        if path:
            self._module_specs[path] = module_spec

        program = self.parser.parse(contents)

        header_processor = HeaderProcessor(self, module_spec, self.include_as)
        for header in program.headers:
            header.apply(header_processor)

        generator = Generator(module_spec.scope, strict=self.strict)
        for definition in program.definitions:
            generator.process(definition)

        return module_spec


class HeaderProcessor(object):
    """Processes headers found in the Thrift file."""

    __slots__ = ('compiler', 'module_spec', 'include_as')

    def __init__(self, compiler, module_spec, include_as):
        self.compiler = compiler
        self.module_spec = module_spec
        self.include_as = include_as

    def visit_include(self, include):

        if not self.module_spec.can_include:
            raise ThriftCompilerError(
                'Include of "%s" found on line %d. '
                'Includes are not supported when using the "loads()" API.'
                'Try loading the file using the "load()" API.'
                % (include.path, include.lineno)
            )

        if not any(include.path.startswith(p) for p in ('./', '../')):
            raise ThriftCompilerError(
                'Paths in include statements are relative to the directory '
                'containing the Thrift file. They must be in the form '
                '"./foo.thrift" or "../bar.thrift".'
            )

        # Includes are relative to directory of the Thrift file being
        # compiled.
        path = os.path.join(
            os.path.dirname(self.module_spec.path), include.path
        )

        if not os.path.isfile(path):
            raise ThriftCompilerError(
                'Cannot include "%s" on line %d in %s. '
                'The file "%s" does not exist.'
                % (include.path, include.lineno, self.module_spec.path, path)
            )

        # name is the name of the module. included_name is the name under
        # which it's exposed inside the current module.
        name = os.path.splitext(os.path.basename(include.path))[0]

        if include.name and not self.include_as:
            raise ThriftCompilerError(
                'Cannot include "%s" as "%s" on line %d. '
                'The "include-as" syntax is currently disabled. '
                'Enable it by instantiating a Loader with include_as=True'
                % (name, include.name, include.lineno)
            )

        included_name = include.name
        if included_name is None:
            included_name = name

        with open(path, 'r') as f:
            contents = f.read()

        included_module_spec = self.compiler.compile(name, contents, path)
        self.module_spec.add_include(included_name, included_module_spec)

    def visit_namespace(self, namespace):
        pass  # nothing to do



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import numbers
import decimal
import fractions

from thriftrw.wire cimport ttype
from thriftrw.protocol.core cimport ProtocolWriter, ProtocolReader
from thriftrw.wire.value cimport (
    Value,
    BoolValue,
    ByteValue,
    DoubleValue,
    I16Value,
    I32Value,
    I64Value,
    BinaryValue,
)

from . cimport check
from .base cimport TypeSpec

__all__ = [
    'BoolTypeSpec',
    'ByteTypeSpec',
    'DoubleTypeSpec',
    'I16TypeSpec',
    'I32TypeSpec',
    'I64TypeSpec',
    'BinaryTypeSpec',
    'TextTypeSpec',
]


cdef class PrimitiveTypeSpec(TypeSpec):
    """TypeSpec for primitive types.

    :param name:
        Name of the primitive type
    :param code:
        TType code used by this primitive
    :param value_cls:
        Constructor for wire value types
    :param surface:
        Values passed through this spec must be instances of this class or
        an exception will be raised
    :param cast:
        If provided, this is used to cast values into a standard shape
        before passing them to ``value_cls``
    :param validate_extra:
        If provided, this is used to validate values beyond checking that
        they have the correct type.
    """

    cpdef Value to_wire(self, object value):
        return self.value_cls(self.cast(value))

    cpdef object to_primitive(PrimitiveTypeSpec self, object value):
        return value

    cpdef object from_wire(PrimitiveTypeSpec self, Value wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value

    cpdef object from_primitive(PrimitiveTypeSpec self, object prim_value):
        return self.cast(prim_value)

    cpdef TypeSpec link(self, scope):
        return self

    cpdef void validate(PrimitiveTypeSpec self, object instance) except *:
        check.instanceof_surface(self, instance)
        if self.validate_extra is not None:
            self.validate_extra(instance)

    def __str__(self):
        return 'PrimitiveTypeSpec(%r, %s)' % (self.code, self.value_cls)

    def __repr__(self):
        return str(self)

    def __richcmp__(PrimitiveTypeSpec self, PrimitiveTypeSpec other, int op):
        if op == 2:
            return self is other
        else:
            return False


cdef class _TextualTypeSpec(TypeSpec):

    def __init__(self):
        self.ttype_code = ttype.BINARY

    cpdef TypeSpec link(self, scope):
        return self

    cpdef Value to_wire(_TextualTypeSpec self, object value):
        if type(value) is unicode:
            value = value.encode('utf-8')
        return BinaryValue(value)

    cpdef void write_to(_TextualTypeSpec self, ProtocolWriter writer,
                        object value) except *:
        if type(value) is unicode:
            value = value.encode('utf-8')
        writer.write_binary(value, len(value))

    cpdef void validate(_TextualTypeSpec self, object instance) except *:
        if not isinstance(instance, (bytes, unicode)):
            raise TypeError(
                'Cannot convert %r into a "%s".' % (instance, self.name)
            )


cdef class _TextTypeSpec(_TextualTypeSpec):
    """TypeSpec for the text type."""

    name = str('string')
    surface = unicode

    cpdef object read_from(_TextTypeSpec self, ProtocolReader reader):
        # TODO: Is this right?
        return reader.read_binary().decode('utf-8')

    cpdef object to_primitive(_TextTypeSpec self, object value):
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        return value

    cpdef object from_wire(_TextTypeSpec self, Value wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value.decode('utf-8')

    cpdef object from_primitive(_TextTypeSpec self, object prim_value):
        if isinstance(prim_value, bytes):
            prim_value = prim_value.decode('utf-8')
        return prim_value

    def __richcmp__(_TextTypeSpec self, _TextTypeSpec other, int op):
        if op == 2:
            return self is other
        else:
            return False


cdef class _BinaryTypeSpec(_TextualTypeSpec):

    name = str('binary')
    surface = bytes

    cpdef object read_from(_BinaryTypeSpec self, ProtocolReader reader):
        return reader.read_binary()

    cpdef object to_primitive(_BinaryTypeSpec self, object value):
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        return value

    cpdef object from_wire(_BinaryTypeSpec self, Value wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value

    cpdef object from_primitive(_BinaryTypeSpec self, object prim_value):
        if isinstance(prim_value, unicode):
            prim_value = prim_value.encode('utf-8')
        return prim_value

    def __richcmp__(_BinaryTypeSpec self, _BinaryTypeSpec other, int op):
        if op == 2:
            return self is other
        else:
            return False


cdef class _BoolTypeSpec(TypeSpec):

    name = str('bool')
    surface = bool

    def __init__(self):
        self.ttype_code = ttype.BOOL

    cpdef object read_from(_BoolTypeSpec self, ProtocolReader reader):
        return reader.read_bool()

    cpdef Value to_wire(_BoolTypeSpec self, object value):
        return BoolValue(bool(value))

    cpdef object to_primitive(_BoolTypeSpec self, object value):
        return bool(value)

    cpdef object from_wire(_BoolTypeSpec self, Value wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value

    cpdef object from_primitive(_BoolTypeSpec self, object prim_value):
        return bool(prim_value)

    cpdef void write_to(_BoolTypeSpec self, ProtocolWriter writer,
                        object value) except *:
        writer.write_bool(value)

    cpdef TypeSpec link(self, scope):
        return self

    cpdef void validate(_BoolTypeSpec self, object instance) except *:
        check.instanceof_class(self, (bool, int), instance)

    def __richcmp__(_BoolTypeSpec self, _BoolTypeSpec other, int op):
        if op == 2:
            return self is other
        else:
            return False


def validate_signed_int(bits):
    max_magnitude = 1 << (bits - 1)
    min_value = -1 * max_magnitude
    max_value = max_magnitude - 1
    def _validate_signed_int(x):
        if x < min_value or x > max_value:
            raise ValueError('Value %d does not fit in an i%d' % (x, bits))
    return _validate_signed_int


# Although the Integral and Number classes already cover int, long, (and
# everything else in case Number), we list those out here explicitly because
# old versions of PyPy or PyPy3 don't play well with isinstance checks
# generated by Cython.
_INTEGRAL = (int, long, numbers.Integral)
_FLOATING = (int, long, float, decimal.Decimal, fractions.Fraction,
             numbers.Number)

BoolTypeSpec = _BoolTypeSpec()

cdef class _ByteTypeSpec(PrimitiveTypeSpec):

    def __init__(self):
        self.name = str('byte')
        self.ttype_code = ttype.BYTE
        self.value_cls = ByteValue
        self.surface = _INTEGRAL
        self.cast = int
        self.validate_extra = validate_signed_int(8)

    cpdef void write_to(_ByteTypeSpec self, ProtocolWriter writer,
                        object value) except *:
        writer.write_byte(value)

    cpdef object read_from(_ByteTypeSpec self, ProtocolReader reader):
        return reader.read_byte()

ByteTypeSpec = _ByteTypeSpec()

cdef class _DoubleTypeSpec(PrimitiveTypeSpec):

    def __init__(self):
        self.name = str('double')
        self.ttype_code = ttype.DOUBLE
        self.value_cls = DoubleValue
        self.surface = _FLOATING
        self.cast = float

    cpdef void write_to(_DoubleTypeSpec self, ProtocolWriter writer,
                        object value) except *:
        writer.write_double(value)

    cpdef object read_from(_DoubleTypeSpec self, ProtocolReader reader):
        return reader.read_double()

DoubleTypeSpec = _DoubleTypeSpec()

cdef class _I16TypeSpec(PrimitiveTypeSpec):

    def __init__(self):
        self.name = str('i16')
        self.ttype_code = ttype.I16
        self.value_cls = I16Value
        self.surface = _INTEGRAL
        self.cast = int
        self.validate_extra = validate_signed_int(16)

    cpdef void write_to(_I16TypeSpec self, ProtocolWriter writer,
                        object value) except *:
        writer.write_i16(value)

    cpdef object read_from(_I16TypeSpec self, ProtocolReader reader):
        return reader.read_i16()

I16TypeSpec = _I16TypeSpec()

cdef class _I32TypeSpec(PrimitiveTypeSpec):

    def __init__(self):
        self.name = str('i32')
        self.ttype_code = ttype.I32
        self.value_cls = I32Value
        self.surface = _INTEGRAL
        self.cast = int
        self.validate_extra = validate_signed_int(32)

    cpdef void write_to(_I32TypeSpec self, ProtocolWriter writer,
                        object value) except *:
        writer.write_i32(value)

    cpdef object read_from(_I32TypeSpec self, ProtocolReader reader):
        return reader.read_i32()

I32TypeSpec = _I32TypeSpec()

cdef class _I64TypeSpec(PrimitiveTypeSpec):

    def __init__(self):
        self.name = str("i64")
        self.ttype_code = ttype.I64
        self.value_cls = I64Value
        self.surface = _INTEGRAL
        self.cast = long
        self.validate_extra = validate_signed_int(64)

    cpdef void write_to(_I64TypeSpec self, ProtocolWriter writer,
                        object value) except *:
        writer.write_i64(value)

    cpdef object read_from(_I64TypeSpec self, ProtocolReader reader):
        return reader.read_i64()

I64TypeSpec = _I64TypeSpec()

BinaryTypeSpec = _BinaryTypeSpec()

TextTypeSpec = _TextTypeSpec()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
.. autoclass:: ThriftError
    :members:

.. autoclass:: ThriftParserError
    :members:

.. autoclass:: ThriftCompilerError
    :members:

.. autoclass:: ThriftProtocolError
    :members:

.. autoclass:: EndOfInputError
    :members:

.. autoclass:: UnknownExceptionError
    :members:
"""
from __future__ import absolute_import, unicode_literals, print_function


class ThriftError(Exception):
    """Base class for all exceptions raised by thriftrw."""


class ThriftParserError(ThriftError):
    """Exception raised by the parser or lexer in case of errors."""


class ThriftCompilerError(ThriftError):
    """Exception raised during IDL compilation."""


class ThriftProtocolError(ThriftError):
    """Exceptions raised by Protocol implementations for errors encountered
    during serialization or deserialization.
    """


class EndOfInputError(ThriftProtocolError):
    """The input was shorter than expected."""


class UnknownExceptionError(ThriftError):
    """We parsed an unknown exception in a function response."""

    def __init__(self, message, thrift_response=None):
        super(UnknownExceptionError, self).__init__(message)
        self.thrift_response = thrift_response

    def __str__(self):
        return 'UnknownExceptionError(%s, thrift_response=%r)' % (
            super(UnknownExceptionError, self).__str__(),
            self.thrift_response,
        )

    __repr__ = __str__



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from libc.stdint cimport (
    int8_t,
    int16_t,
    int32_t,
    int64_t,
)

from . cimport ttype

from thriftrw._cython cimport richcompare

__all__ = [
    'Value',
    'BoolValue',
    'ByteValue',
    'DoubleValue',
    'I16Value',
    'I32Value',
    'I64Value',
    'BinaryValue',
    'FieldValue',
    'StructValue',
    'MapValue',
    'MapItem',
    'SetValue',
    'ListValue',
    'ValueVisitor',
]


cdef class Value(object):
    """Base class for Value classes.

    Value classes define an intermediate representation of Thrift types as
    they are sent and received over the wire.
    """

    cpdef object apply(Value self, ValueVisitor visitor):
        """Apply the value to the given visitor.

        The appropriate `visit_*` method will be called and its result
        returned.

        :param ValueVisitor visitor:
            Visitor on the value.
        :returns:
            Value returned by the corresponding ``visit_*`` method.
        """
        raise NotImplementedError


cdef class BoolValue(Value):
    """Wrapper for boolean values."""

    ttype_code = ttype.BOOL

    def __cinit__(BoolValue self, bint value):
        self.value = value

    def __richcmp__(BoolValue self, BoolValue other not None, int op):
        return richcompare(op, [(self.value, other.value)])

    def __str__(self):
        return 'BoolValue(%r)' % (self.value,)

    def __repr__(self):
        return str(self)

    cpdef object apply(BoolValue self, ValueVisitor visitor):
        return visitor.visit_bool(self.value)


cdef class ByteValue(Value):
    """Wrapper for byte values."""

    ttype_code = ttype.BYTE

    def __cinit__(ByteValue self, int8_t value):
        self.value = value

    def __richcmp__(ByteValue self, ByteValue other not None, int op):
        return richcompare(op, [(self.value, other.value)])

    def __str__(self):
        return 'ByteValue(%r)' % (self.value,)

    def __repr__(self):
        return str(self)

    cpdef object apply(ByteValue self, ValueVisitor visitor):
        return visitor.visit_byte(self.value)


cdef class DoubleValue(Value):
    """Wrapper for double values."""

    ttype_code = ttype.DOUBLE

    def __cinit__(DoubleValue self, double value):
        self.value = value

    def __richcmp__(DoubleValue self, DoubleValue other not None, int op):
        return richcompare(op, [(self.value, other.value)])

    def __str__(self):
        return 'DoubleValue(%r)' % (self.value,)

    def __repr__(self):
        return str(self)

    cpdef object apply(DoubleValue self, ValueVisitor visitor):
        return visitor.visit_double(self.value)


cdef class I16Value(Value):
    """Wrapper for 16-bit integer values."""

    ttype_code = ttype.I16

    def __cinit__(I16Value self, int16_t value):
        self.value = value

    def __richcmp__(I16Value self, I16Value other not None, int op):
        return richcompare(op, [(self.value, other.value)])

    def __str__(self):
        return 'I16Value(%r)' % (self.value,)

    def __repr__(self):
        return str(self)

    cpdef object apply(I16Value self, ValueVisitor visitor):
        return visitor.visit_i16(self.value)


cdef class I32Value(Value):
    """Wrapper for 32-bit integer values."""

    ttype_code = ttype.I32

    def __cinit__(I32Value self, int value):
        self.value = value

    def __richcmp__(I32Value self, I32Value other not None, int op):
        return richcompare(op, [(self.value, other.value)])

    def __str__(self):
        return 'I32Value(%r)' % (self.value,)

    def __repr__(self):
        return str(self)

    cpdef object apply(I32Value self, ValueVisitor visitor):
        return visitor.visit_i32(self.value)


cdef class I64Value(Value):
    """Wrapper for 64-bit integer values."""

    ttype_code = ttype.I64

    def __cinit__(I64Value self, long value):
        self.value = value

    def __richcmp__(I64Value self, I64Value other not None, int op):
        return richcompare(op, [(self.value, other.value)])

    def __str__(self):
        return 'I64Value(%r)' % (self.value,)

    def __repr__(self):
        return str(self)

    cpdef object apply(I64Value self, ValueVisitor visitor):
        return visitor.visit_i64(self.value)


cdef class BinaryValue(Value):
    """Wrapper for binary blobs.

    Note that Thrift does not differentiate between text and binary blobs over
    the wire. UTF-8 text should be encoded/decoded manually.
    """

    ttype_code = ttype.BINARY

    def __cinit__(BinaryValue self, bytes value):
        self.value = value

    def __richcmp__(BinaryValue self, BinaryValue other not None, int op):
        return richcompare(op, [(self.value, other.value)])

    def __str__(self):
        return 'BinaryValue(%r)' % (self.value,)

    def __repr__(self):
        return str(self)

    cpdef object apply(BinaryValue self, ValueVisitor visitor):
        return visitor.visit_binary(self.value)


cdef class FieldValue(object):
    """A single field in a struct.

    .. py:attribute:: id

        Field identifier.

    .. py:attribute:: ttype

        :py:data:`~thriftrw.wire.TType` of the value held in this field.

    .. py:attribute:: value

        Value for this field.
    """

    def __cinit__(FieldValue self, int16_t id, int8_t ttype, value):
        self.id = id
        self.ttype = ttype
        self.value = value

    def __str__(self):
        return 'FieldValue(%r, %r, %r)' % (self.id, self.ttype, self.value)

    def __repr__(self):
        return str(self)

    def __richcmp__(FieldValue self, FieldValue other not None, int op):
        return richcompare(op, [
            (self.id, other.id),
            (self.ttype, other.ttype),
            (self.value, other.value),
        ])


cdef class StructValue(Value):
    """A struct value is a collection of fields of different types.

    .. py:attribute:: fields

        Collection of :py:class:`FieldValue` objects.
    """

    ttype_code = ttype.STRUCT

    def __cinit__(StructValue self, list fields):
        self.fields = fields

    def __init__(self, list fields):
        self._index = {}
        for field in fields:
            self._index[(field.id, field.ttype)] = field

    cpdef object apply(StructValue self, ValueVisitor visitor):
        return visitor.visit_struct(self.fields)

    def get(self, field_id, field_ttype):
        """Returns the value at the given field ID and type.

        :param field_id:
            Numerical field identifier.
        :param field_ttype:
            Type of the value.
        :returns:
            Corresponding ``FieldValue`` or None.
        """
        field_value = self._index.get((field_id, field_ttype))
        if field_value is None:
            return None
        else:
            return field_value

    def __richcmp__(StructValue self, StructValue other not None, int op):
        return richcompare(op, [(self.fields, other.fields)])

    def __str__(self):
        return 'StructValue(%r)' % self.fields

    def __repr__(self):
        return str(self)


cdef class MapItem(object):
    """An item in a map.

    .. py:attribute:: key

        Key of the item

    .. py:attribute:: value

        Value associated with the key
    """

    def __cinit__(MapItem self, key, value):
        self.key = key
        self.value = value

    def __str__(self):
        return 'MapValue(%r, %r)' % (self.key, self.value)

    def __repr__(self):
        return str(self)

    def __richcmp__(MapItem self, MapItem other not None, int op):
        return richcompare(op, [
            (self.key, other.key),
            (self.value, other.value),
        ])


cdef class MapValue(Value):
    """A mapping of two different kinds of values.

    This object may be treated as a map.

    .. py:attribute:: key_ttype

        Type of the keys stored in the map. This must be a value from TType.

    .. py:attribute:: value_ttype

        Type of the values stored in the map. This must be a value from TType.

    .. py:attribute:: pairs

        Collection of :py:class:`thriftrw.wire.MapItem` objects.
    """

    ttype_code = ttype.MAP

    def __cinit__(MapValue self, int8_t key_ttype, int8_t value_ttype, list pairs):
        self.key_ttype = key_ttype
        self.value_ttype = value_ttype
        self.pairs = pairs

    def __richcmp__(MapValue self, MapValue other not None, int op):
        return richcompare(op, [
            (self.key_ttype, other.key_ttype),
            (self.value_ttype, other.value_ttype),
            (self.pairs, other.pairs),
        ])

    def __str__(self):
        return 'MapValue(%r)' % (self.pairs,)

    def __repr__(self):
        return str(self)

    cpdef object apply(MapValue self, ValueVisitor visitor):
        return visitor.visit_map(self.key_ttype, self.value_ttype, self.pairs)


cdef class SetValue(Value):
    """A collection of unique values of the same type.

    .. py:attribute:: value_ttype

        Type of values in the set. This must be a value from TType.

    .. py:attribute:: values

        Collection of the values.
    """

    ttype_code = ttype.SET

    def __cinit__(SetValue self, int8_t value_ttype, list values):
        self.value_ttype = value_ttype
        self.values = values

    def __str__(self):
        return 'SetValue(%r)' % (self.values,)

    def __repr__(self):
        return str(self)

    def __richcmp__(SetValue self, SetValue other not None, int op):
        return richcompare(op, [
            (self.value_ttype, other.value_ttype),
            (self.values, other.values),
        ])

    cpdef object apply(SetValue self, ValueVisitor visitor):
        return visitor.visit_set(self.value_ttype, self.values)


cdef class ListValue(Value):
    """A collection of values.

    .. py:attribute:: value_ttype

        Type of values in the list. This must be a value from TType.

    .. py:attribute:: values

        Collection of the values.
    """

    ttype_code = ttype.LIST

    def __cinit__(ListValue self, int8_t value_ttype, list values):
        self.value_ttype = value_ttype
        self.values = values

    def __str__(self):
        return 'ListValue(%r)' % (self.values,)

    def __repr__(self):
        return str(self)

    def __richcmp__(ListValue self, ListValue other not None, int op):
        return richcompare(op, [
            (self.value_ttype, other.value_ttype),
            (self.values, other.values),
        ])

    cpdef object apply(ListValue self, ValueVisitor visitor):
        return visitor.visit_list(self.value_ttype, self.values)


cdef class ValueVisitor(object):
    """Visitor on different value types.

    The ``visit_*`` functions are not given the ``*Value`` objects but the
    actual values contained in them.

    The idea is that when something needs to take an arbitrary Thrift value
    and act on all cases, it extends this type. Each Thrift value knows which
    function to call on the visitor based on the value type. The intention
    here is to avoid ``isinstance`` checks.
    """

    cdef object visit_bool(self, bint value):
        """Visits boolean values.

        :param bool value:
            True or False
        """
        raise NotImplementedError

    cdef object visit_byte(self, int8_t value):
        """Visits 8-bit integers.

        :param int value:
            8-bit integer
        """
        raise NotImplementedError

    cdef object visit_double(self, double value):
        """Visits double values.

        :param float value:
            Floating point number
        """
        raise NotImplementedError

    cdef object visit_i16(self, int16_t value):
        """Visits 16-bit integers.

        :param int value:
            16-bit integer
        """
        raise NotImplementedError

    cdef object visit_i32(self, int32_t value):
        """Visits 32-bit integers.

        :param int value:
            32-bit integer
        """
        raise NotImplementedError

    cdef object visit_i64(self, int64_t value):
        """Visits 64-bit integers.

        :param int value:
            64-bit integer
        """
        raise NotImplementedError

    cdef object visit_binary(self, bytes value):
        """Visits binary blobs.

        :param bytes value:
            Binary blob
        """
        raise NotImplementedError

    cdef object visit_struct(self, list fields):
        """Visits structs.

        :param fields:
            Collection of :py:class:`FieldValue` objects.
        """
        raise NotImplementedError

    cdef object visit_map(
        self, int8_t key_ttype, int8_t value_ttype, list pairs
    ):
        """Visits maps.

        :param thriftrw.wire.TType key_ttype:
            TType of the keys in the map.
        :param thriftrw.wire.TType value_ttype:
            TType of the values in the map.
        :param pairs:
            Collection of key-value pairs.
        """
        raise NotImplementedError

    cdef object visit_set(self, int8_t value_ttype, list values):
        """Visits sets.

        :param thriftrw.wire.TType value_ttype:
            TType of the items in the set.
        :param values:
            Collection of values in the set.
        """
        raise NotImplementedError

    cdef object visit_list(self, int8_t value_ttype, list values):
        """Visits lists.

        :param thriftrw.wire.TType value_ttype:
            TType of the items in the set.
        :param values:
            Collection of values in the list.
        """
        raise NotImplementedError


cdef class _ValueVisitorArgs(ValueVisitor):
    """A ValueVisitor that simply returns the arguments for a visitor along
    with its name.

    This is meant for testing only.
    """

    cdef object visit_bool(self, bint value):
        return ('visit_bool', value)

    cdef object visit_byte(self, int8_t value):
        return ('visit_byte', value)

    cdef object visit_double(self, double value):
        return ('visit_double', value)

    cdef object visit_i16(self, int16_t value):
        return ('visit_i16', value)

    cdef object visit_i32(self, int32_t value):
        return ('visit_i32', value)

    cdef object visit_i64(self, int64_t value):
        return ('visit_i64', value)

    cdef object visit_binary(self, bytes value):
        return ('visit_binary', value)

    cdef object visit_struct(self, list fields):
        return ('visit_struct', fields)

    cdef object visit_map(
        self, int8_t key_ttype, int8_t value_ttype, list pairs
    ):
        return ('visit_map', key_ttype, value_ttype, pairs)

    cdef object visit_set(self, int8_t value_ttype, list values):
        return ('visit_set', value_ttype, values)

    cdef object visit_list(self, int8_t value_ttype, list values):
        return ('visit_list', value_ttype, values)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Provides functionality that is used exclusively at runtime.
"""
from __future__ import absolute_import, unicode_literals, print_function

from libc.stdint cimport int32_t

from thriftrw._buffer cimport WriteBuffer, ReadBuffer
from thriftrw.protocol.core cimport (
    Protocol,
    ProtocolWriter,
    ProtocolReader,
    MessageHeader,
)
from thriftrw.wire cimport mtype
from thriftrw.wire cimport ttype
from thriftrw.wire.value cimport Value, StructValue
from thriftrw.spec.service import (
    FunctionArgsSpec,
    FunctionResultSpec,
)
from thriftrw.errors import (
    ThriftProtocolError,
    UnknownExceptionError,
)

cdef class Serializer(object):
    def __cinit__(self, Protocol protocol):
        self.protocol = protocol

    def __call__(self, obj):
        """Serializes the given Thrift struct.

        :param obj:
            An instance of a class generated by thriftrw representing a
            struct, union, or exception.
        :returns:
            Binary representation of the object.
        """
        return self.dumps(obj)

    cpdef bytes dumps(self, obj):
        cdef WriteBuffer buff = WriteBuffer()
        cdef ProtocolWriter writer = self.protocol.writer(buff)

        obj.__class__.type_spec.write_to(writer, obj)
        return buff.value

    cpdef bytes message(self, obj, int32_t seqid=0):
        """Serializes the given request or response into a Thrift Message.

        :param obj:
            Request or response to serialize. This is constructed using the
            ``request`` or ``response`` attribute on a
            :py:class:`thriftrw.spec.ServiceFunction`.
        :param int seqid:
            If given, this specifies the seqid to use for the message.
            Defaults to 0.
        """
        obj_spec = obj.__class__.type_spec

        if isinstance(obj_spec, FunctionArgsSpec):
            function_spec = obj_spec.function
            if function_spec.oneway:
                message_type = mtype.ONEWAY
            else:
                message_type = mtype.CALL
        elif isinstance(obj_spec, FunctionResultSpec):
            function_spec = obj_spec.function
            message_type = mtype.REPLY
        else:
            raise TypeError(
                'Only function request or response types may be wrapped '
                'in messages.'
            )

        cdef bytes name
        if isinstance(function_spec.name, unicode):
            name = function_spec.name.encode('utf-8')
        else:
            name = function_spec.name

        cdef WriteBuffer buff = WriteBuffer()
        cdef ProtocolWriter writer = self.protocol.writer(buff)
        cdef MessageHeader header = MessageHeader(name, message_type, seqid)

        writer.write_message_begin(header)
        obj_spec.write_to(writer, obj)
        writer.write_message_end()

        return buff.value

cdef class Deserializer(object):
    def __cinit__(self, Protocol protocol):
        self.protocol = protocol

    def __call__(self, obj_cls, bytes s):
        """Deserializes an object from the given blob.

        :param obj_cls:
            A class generated by thriftrw representing a struct, union, or
            exception.
        :param bytes s:
            Binary blob representing the object.
        :returns:
            Deserialized object.
        :raises thriftrw.errors.ThriftProtocolError:
            If the object failed to deserialize.
        """
        return self.loads(obj_cls, s)

    cpdef object loads(self, obj_cls, bytes s):
        cdef ReadBuffer buff = ReadBuffer(s)
        cdef ProtocolReader reader = self.protocol.reader(buff)
        return obj_cls.type_spec.read_from(reader)

    cpdef Message message(self, service, bytes s):
        """Deserializes a message from the given blob.

        :param service:
            Reference to a service class. The request or response for one of
            the methods of this service will be read from the blob based on
            the message type.
        :param s:
            Binary blob representing the message and its payload.
        :returns:
            A Message containing the parsed request or response object in the
            ``body``.
        :raises UnknownExceptionError:
            If a ``EXCEPTION`` message was parsed.
        :raises ThriftProtocolError:
            If the method name is not recognized or if any other parsing error
            occurs.
        """
        cdef ProtocolReader reader = self.protocol.reader(ReadBuffer(s))
        cdef Message message
        cdef object body

        cdef MessageHeader header = reader.read_message_begin()

        # Use message header to find a spec for deserializing body.

        if header.type == mtype.EXCEPTION:
            # For EXCEPTION messages, just raise UnknownExceptionError with
            # the struct representation in the message.
            raise UnknownExceptionError('Received an exception message.', header)

        function_spec = service.service_spec.lookup(header.name)
        if function_spec is None:
            raise ThriftProtocolError(
                'Unknown method "%s" referenced by message %r'
                % (header.name, header)
            )

        if (
            header.type == mtype.CALL or
            header.type == mtype.ONEWAY
        ):
            body = function_spec.args_spec.read_from(reader)
        elif header.type == mtype.REPLY:
            if function_spec.oneway:
                raise ThriftProtocolError(
                    'Function "%s" is a oneway method. '
                    'It cannot receive a REPLY.' % function_spec.name
                )

            body = function_spec.result_spec.read_from(reader)
        else:
            # Unrecognized message type. If this happens, we have a bug
            # because deserialize_message already raises an exception for
            # invalid message IDs.
            raise ValueError(
                'Unknown message type %d in message %r'
                % (header.type, header)
            )

        reader.read_message_end()

        return Message(header.name, header.seqid, header.type, body)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from . cimport check
from .base cimport TypeSpec
from thriftrw.wire cimport ttype
from thriftrw._cython cimport richcompare
from thriftrw.wire.value cimport MapItem, MapValue, Value
from thriftrw.protocol.core cimport MapHeader, ProtocolWriter, ProtocolReader

__all__ = ['MapTypeSpec']


cdef class MapTypeSpec(TypeSpec):
    """Spec for map types.

    .. py:attribute:: kspec

        TypeSpec for the kind of keys matching maps must contain.

    .. py:attribute:: vspec

        TypeSpec for the kind of values matching maps must contain.
    """

    surface = dict

    def __init__(self, kspec, vspec):
        """
        :param TypeSpec kspec:
            TypeSpec of the keys in the map.
        :param TypeSpec vspec:
            TypeSpec of the values in the map.
        """
        self.ttype_code = ttype.MAP
        self.kspec = kspec
        self.vspec = vspec
        self.linked = False

    @property
    def name(self):
        return 'map<%s, %s>' % (self.kspec.name, self.vspec.name)

    cpdef TypeSpec link(self, scope):
        if not self.linked:
            self.linked = True
            self.kspec = self.kspec.link(scope)
            self.vspec = self.vspec.link(scope)
        return self

    cpdef Value to_wire(MapTypeSpec self, object value):
        return MapValue(
            key_ttype=self.kspec.ttype_code,
            value_ttype=self.vspec.ttype_code,
            pairs=[
                MapItem(self.kspec.to_wire(k), self.vspec.to_wire(v))
                for k, v in value.items()
            ]
        )

    cpdef object to_primitive(MapTypeSpec self, object value):
        return {
            self.kspec.to_primitive(k): self.vspec.to_primitive(v)
            for k, v in value.items()
        }

    cpdef object read_from(MapTypeSpec self, ProtocolReader reader):
        cdef MapHeader header = reader.read_map_begin()
        cdef dict output = {
            self.kspec.read_from(reader): self.vspec.read_from(reader)
            for i in range(header.size)
        }
        reader.read_map_end()
        return output

    cpdef void write_to(MapTypeSpec self, ProtocolWriter writer,
                        object value) except *:
        cdef MapHeader header = MapHeader(
            self.kspec.ttype_code, self.vspec.ttype_code, len(value)
        )
        writer.write_map_begin(header)
        for k, v in value.items():
            self.kspec.write_to(writer, k)
            self.vspec.write_to(writer, v)
        writer.write_map_end()

    cpdef object from_wire(MapTypeSpec self, Value wire_value):
        check.type_code_matches(self, wire_value)
        return {
            self.kspec.from_wire(i.key): self.vspec.from_wire(i.value)
            for i in wire_value.pairs
        }

    cpdef object from_primitive(MapTypeSpec self, object prim_value):
        return {
            self.kspec.from_primitive(k): self.vspec.from_primitive(v)
            for k, v in prim_value.items()
        }

    cpdef void validate(MapTypeSpec self, object instance) except *:
        check.ismapping(self, instance)
        for k, v in instance.items():
            self.kspec.validate(k)
            self.vspec.validate(v)

    def __str__(self):
        return 'MapTypeSpec(kspec=%r, vspec=%r)' % (self.kspec, self.vspec)

    def __repr__(self):
        return str(self)

    def __richcmp__(MapTypeSpec self, MapTypeSpec other not None, int op):
        return richcompare(op, [
            (self.kspec, other.kspec),
            (self.vspec, other.vspec),
        ])




# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from collections import namedtuple

from thriftrw.wire.value cimport Value
from thriftrw.protocol.core cimport ProtocolReader, FieldHeader
from .field cimport FieldSpec
from .union cimport UnionTypeSpec
from .struct cimport StructTypeSpec
from . cimport check

from .spec_mapper import type_spec_or_ref
from ..errors import ThriftCompilerError
from ..errors import UnknownExceptionError

__all__ = [
    'ServiceSpec',
    'FunctionSpec',
    'ServiceFunction',
    'FunctionArgsSpec',
    'FunctionResultSpec',
]


cdef class FunctionArgsSpec(StructTypeSpec):
    """Represents the parameters of a service function.

    .. py:attribute:: function

        The :py:class:`FunctionSpec` whose arguments this spec represents.
        This value is available only after the spec has been linked.

    The parameters of a function implicitly form a struct which contains the
    parameters as its fields, which are optional by default.

    The surface for this is the same as :py:class:`StructTypeSpec` except that
    the generated class also includes the following class attributes.

    .. py:attribute:: result_type

        A reference to the class representing the result type for
        ``function``, or None if the function was oneway.

    .. versionchanged:: 1.0

        Added the ``function`` attribute.

    .. versionchanged:: 1.1

        Added ``result_type`` class attribute to the surface.
    """

    def __init__(self, name, params):
        super(FunctionArgsSpec, self).__init__(name, params)
        self.function = None

    @classmethod
    def compile(cls, parameters, service_name, function_name):
        """Compiles a parameter list into a FunctionArgsSpec.

        :param parameters:
            Collection of :py:class:`thriftrw.idl.Field` objects.
        :param str service_name:
            Name of the service under which the function was defined.
        :param str function_name:
            Name of the function whose parameter list is represented by this
            object.
        """
        args_name = str('%s_%s_request' % (service_name, function_name))
        param_specs = [
            FieldSpec.compile(
                field=param,
                struct_name=args_name,
                require_requiredness=False,
            ) for param in parameters
        ]

        return cls(args_name, param_specs)

    def link(self, scope, function):
        self.function = function
        spec = super(FunctionArgsSpec, self).link(scope)

        result_type = None
        if self.function.result_spec is not None:
            result_spec = self.function.result_spec.link(scope, function)
            result_type = result_spec.surface

        spec.surface.result_type = result_type
        return spec


cdef class FunctionResultSpec(UnionTypeSpec):
    """Represents the result of a service function.

    .. py:attribute:: return_spec

        :py:class:`thriftrw.spec.TypeSpec` of the return type or None if the
        function does not return anything.

    .. py:attribute:: exception_specs

        Collection of :py:class:`thriftrw.spec.FieldSpec` objects defining the
        exceptions that this function can raise.

    .. py:attribute:: function

        The :py:class:`FunctionSpec` whose result type this spec represents.
        This value is available only after the spec has been linked.

    The return value of a function and the exceptions raised by it implicitly
    form a union which contains the return value at field ID ``0`` and the
    exceptions on the remaining field IDs.

    .. versionchanged:: 0.2

        Expose the class and add the ``return_spec`` and ``exception_specs``
        attributes.

    .. versionchanged:: 0.5

        When deserializing, if an unrecognized exception is found, a
        :py:class:`thriftrw.errors.UnknownExceptionError` is raised.

    .. versionchanged:: 1.0

        Added the ``function`` attribute.
    """

    def __init__(self, name, return_spec, exceptions):
        self.return_spec = return_spec
        self.exception_specs = exceptions
        self.function = None

        result_specs = []
        if return_spec is not None:
            result_specs.append(
                FieldSpec(
                    id=0,
                    name='success',
                    spec=return_spec,
                    required=False,
                    default_value=None
                )
            )

        result_specs.extend(exceptions)

        # Set of field IDs for exceptions recognized by this result. If the
        # system tries to read a field with an unrecognized exception ID, an
        # ``UnknownExceptionError`` will be raised.
        self.exception_ids = frozenset([f.id for f in exceptions])

        super(FunctionResultSpec, self).__init__(
            name, result_specs, allow_empty=(return_spec is None)
        )

    cpdef object read_from(FunctionResultSpec self, ProtocolReader reader):
        reader.read_struct_begin()

        cdef dict kwargs = {}
        cdef object val
        cdef FieldSpec spec
        cdef FieldHeader header

        header = reader.read_field_begin()

        # We use -1 to signify struct end due to cython constraints.
        while header.type != -1:
            spec = self._index.get((header.id, header.type), None)

            if spec is None:
                if header.id != 0:
                    raise UnknownExceptionError(
                        (
                            '"%s" received an unrecognized exception. '
                            'Make sure your Thrift IDL is up-to-date with '
                            'what the remote host is using.'
                        ) % self.name,
                    )
                else:
                    reader.skip(header.type)
            else:
                val = spec.spec.read_from(reader)
                kwargs[spec.name] = val

            reader.read_field_end()
            header = reader.read_field_begin()

        reader.read_struct_end()

        return self.surface(**kwargs)

    cpdef object from_wire(self, Value wire_value):
        check.type_code_matches(self, wire_value)

        for field in wire_value.fields:
            if field.id != 0 and field.id not in self.exception_ids:
                raise UnknownExceptionError(
                    (
                        '"%s" received an unrecognized exception. '
                        'Make sure your Thrift IDL is up-to-date with '
                        'what the remote host is using.'
                    ) % self.name,
                    wire_value,
                )

        return super(FunctionResultSpec, self).from_wire(wire_value)

    def link(self, scope, function):
        self.function = function
        if self.return_spec is not None:
            self.return_spec = self.return_spec.link(scope)
        self.exception_specs = [e.link(scope) for e in self.exception_specs]
        return super(FunctionResultSpec, self).link(scope)

    @classmethod
    def compile(cls, return_type, exceptions, service_name, function_name):
        """Compiles information from the AST into a FunctionResultSpec.

        :param return_type:
            A :py:class:`thriftrw.idl.Type` representing the return type or
            None if the function doesn't return anything.
        :param exceptions:
            Collection of :py:class:`thriftrw.idl.Field` objects representing
            raised by the function.
        :param str service_name:
            Name of the service under which the function was defined.
        :param str function_name:
            Name of the function whose result this object represents.
        """
        exceptions = exceptions or []
        result_name = str('%s_%s_response' % (service_name, function_name))

        if return_type is not None:
            return_spec = type_spec_or_ref(return_type)
        else:
            return_spec = None

        exc_specs = []
        for exc in exceptions:
            exc_specs.append(
                FieldSpec.compile(
                    field=exc,
                    struct_name=result_name,
                    require_requiredness=False,
                )
            )

        return cls(result_name, return_spec, exc_specs)


cdef class FunctionSpec(object):
    """Specification of a single function on a service.

    .. py:attribute:: name

        Name of the function.

    .. py:attribute:: service

        The :py:class:`ServiceSpec` that contains this function.
        This value is available only after the spec has been linked.

        .. versionadded:: 1.1

    .. py:attribute:: args_spec

        :py:class:`FunctionArgsSpec` specifying the arguments accepted by this
        function as a struct.

    .. py:attribute:: result_spec

        :py:class:`FunctionResultSpec` specifying the output of this
        function as a  union of the return type and the exceptions raised
        by this function.

        The return type of the function (if any) is a field in the union
        with field ID 0 and name 'success'.

        This value is None if the function is oneway.

    .. py:attribute:: oneway

        Whether this function is oneway or not.

    The ``surface`` for a FunctionSpec is a :py:class:`ServiceFunction`
    object. Unlike the ``surface`` for other specs, a FunctionSpec's surface
    is attached to the service class as a class attribute.

    .. versionchanged:: 0.3

        Added the ``oneway`` attribute.
    """

    def __init__(self, name, args_spec, result_spec, oneway):
        self.name = str(name)
        self.args_spec = args_spec
        self.result_spec = result_spec
        self.oneway = oneway
        self.linked = False
        self.surface = None
        self.service = None

    @classmethod
    def compile(cls, func, service_name):
        args_spec = FunctionArgsSpec.compile(
            parameters=func.parameters,
            service_name=service_name,
            function_name=func.name,
        )

        if func.oneway:
            result_spec = None
            if func.return_type is not None:
                raise ThriftCompilerError(
                    'Function "%s.%s" is oneway. It cannot return a value.'
                    % (service_name, func.name)
                )
            if func.exceptions:
                raise ThriftCompilerError(
                    'Function "%s.%s" is oneway. It cannot raise exceptions.'
                    % (service_name, func.name)
                )
        else:
            result_spec = FunctionResultSpec.compile(
                return_type=func.return_type,
                exceptions=func.exceptions,
                service_name=service_name,
                function_name=func.name,
            )

        return cls(func.name, args_spec, result_spec, func.oneway)

    cpdef FunctionSpec link(self, scope, ServiceSpec service):
        if not self.linked:
            self.service = service
            self.linked = True
            if self.result_spec:
                self.result_spec = self.result_spec.link(scope, self)
                result_spec_surface = self.result_spec.surface
            else:
                result_spec_surface = None
            self.args_spec = self.args_spec.link(scope, self)

            self.surface = ServiceFunction(
                self.name,
                self.args_spec.surface,
                result_spec_surface,
                self,
            )
        return self

    def __str__(self):
        return 'FunctionSpec(name=%r, args_spec=%r, result_spec=%r)' % (
            self.name, self.args_spec, self.result_spec
        )

    def __repr__(self):
        return str(self)


cdef class ServiceSpec(object):
    """Spec for a single service.

    .. py:attribute:: name

        Name of the service.

    .. py:attribute:: functions

        Collection of :py:class:`FunctionSpec` objects.

    .. py:attribute:: parent

        ``ServiceSpec`` of the parent service or None if this service does not
        inherit any service.

    The ``surface`` for a ``ServiceSpec`` is a class that has the following
    attributes:

    .. py:attribute:: service_spec

        :py:class:`ServiceSpec` for this service.

    And a reference to one :py:class:`ServiceFunction` object for each
    function defined in the service.
    """

    def __init__(self, name, functions, parent):
        self.name = str(name)
        self.functions = functions
        self.parent = parent

        # For quick function lookups.
        self._functions = None
        self.linked = False
        self.surface = None

    @classmethod
    def compile(cls, service):
        functions = []
        names = set()

        for func in service.functions:
            if func.name in names:
                raise ThriftCompilerError(
                    'Function "%s.%s" cannot be defined. '
                    'That name is already taken.'
                    % (service.name, func.name)
                )
            names.add(func.name)
            functions.append(
                FunctionSpec.compile(func, service.name)
            )

        return cls(service.name, functions, service.parent)

    cpdef ServiceSpec link(self, scope):
        if not self.linked:
            self.linked = True

            if self.parent is not None:
                self.parent = scope.resolve_service_spec(
                    self.parent.name, self.parent.lineno
                )

            self.functions = [func.link(scope, self) for func in self.functions]
            self.surface = service_cls(self, scope)

            # Build index for lookup. Normalize type of names to bytes since
            # it may be bytes or unicode.
            self._functions = {}
            for f in self.functions:
                name = f.name
                if isinstance(name, unicode):
                    name = name.encode('utf-8')
                self._functions[name] = f

        return self

    def lookup(self, name):
        """Look up a function with the given name.

        Returns the function spec or None if no such function exists.

        .. versionadded:: 1.0
        """
        if isinstance(name, unicode):
            name = name.encode('utf-8')
        return self._functions.get(name, None)

    def __str__(self):
        return 'ServiceSpec(name=%r, functions=%r, parent=%r)' % (
            self.name, self.functions, self.parent
        )

    def __repr__(self):
        return str(self)


class ServiceFunction(
    namedtuple('ServiceFunction', 'name request response spec')
):
    """Represents a single function on a service.

    .. py:attribute:: name

        Name of the function.

    .. py:attribute:: request

        Class representing requests for this function.

    .. py:attribute:: response

        Class representing responses for this function, or ``None`` if this
        function is ``oneway``.

    .. py:attribute:: spec

        :py:class:`thriftrw.spec.FunctionSpec` for this function.

    .. versionchanged:: 0.3

        Added the ``spec`` attribute.
    """


def service_cls(service_spec, scope):
    """Generates a class from the given service spec.

    :param ServiceSpec service_spec:
        Specification of the service.
    :param scope:
        Compilation scope.
    """
    parent_cls = object
    if service_spec.parent is not None:
        parent_cls = service_spec.parent.surface

    service_dct = {}
    for function in service_spec.functions:
        service_dct[function.name] = function.surface

    service_dct['service_spec'] = service_spec
    service_dct['__slots__'] = ()

    return type(str(service_spec.name), (parent_cls,), service_dct)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from . cimport check
from .base cimport TypeSpec
from thriftrw.wire cimport ttype
from thriftrw._cython cimport richcompare
from thriftrw.wire.value cimport ListValue
from thriftrw.wire.value cimport Value
from thriftrw.protocol.core cimport ListHeader, ProtocolWriter, ProtocolReader

__all__ = ['ListTypeSpec']


cdef class ListTypeSpec(TypeSpec):
    """Spec for list types.

    .. py:attribute:: vspec

        TypeSpec for the kind of values lists conforming to this spec must
        contain.
    """

    surface = list

    def __init__(self, vspec):
        """
        :param TypeSpec vspec:
            TypeSpec of values stored in the list.
        """
        self.ttype_code = ttype.LIST
        self.vspec = vspec
        self.linked = False

    cpdef TypeSpec link(self, scope):
        if not self.linked:
            self.linked = True
            self.vspec = self.vspec.link(scope)
        return self

    @property
    def name(self):
        return 'list<%s>' % self.vspec.name

    cpdef object read_from(ListTypeSpec self, ProtocolReader reader):
        cdef ListHeader header = reader.read_list_begin()
        cdef list output = []
        for i in range(header.size):
            output.append(self.vspec.read_from(reader))

        reader.read_list_end()
        return output

    cpdef Value to_wire(ListTypeSpec self, object value):
        return ListValue(
            value_ttype=self.vspec.ttype_code,
            values=[self.vspec.to_wire(v) for v in value],
        )

    cpdef object to_primitive(ListTypeSpec self, object value):
        return [self.vspec.to_primitive(x) for x in value]

    cpdef void write_to(ListTypeSpec self, ProtocolWriter writer,
                        object value) except *:
        cdef ListHeader header = ListHeader(self.vspec.ttype_code, len(value))
        writer.write_list_begin(header)
        for v in value:
            self.vspec.write_to(writer, v)
        writer.write_list_end()

    cpdef object from_wire(ListTypeSpec self, Value wire_value):
        check.type_code_matches(self, wire_value)
        return [self.vspec.from_wire(v) for v in wire_value.values]

    cpdef object from_primitive(ListTypeSpec self, object prim_value):
        return [self.vspec.from_primitive(v) for v in prim_value]

    cpdef void validate(ListTypeSpec self, object instance) except *:
        check.isiterable(self, instance)
        for v in instance:
            self.vspec.validate(v)

    def __str__(self):
        return 'ListTypeSpec(vspec=%r)' % self.vspec

    def __repr__(self):
        return str(self)

    def __richcmp__(ListTypeSpec self, ListTypeSpec other not None, int op):
        return richcompare(op, [
            (self.vspec, other.vspec),
        ])




# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from libc.stdint cimport (
    int8_t,
    int16_t,
    int32_t,
    int64_t,
)

from thriftrw.wire cimport ttype
from thriftrw.wire.value cimport Value, ValueVisitor
from thriftrw.wire.message cimport Message


__all__ = ['Protocol']


cdef class MessageHeader(object):

    def __cinit__(MessageHeader self, bytes name, int8_t type, int32_t seqid):
        self.name = name
        self.type = type
        self.seqid = seqid

    def __str__(self):
        return 'MessageHeader(%r, %r, %r)' % (
            self.name, self.seqid, ttype.name_of(self.type)
        )

    def __repr__(self):
        return str(self)


cdef class ProtocolWriter(object):

    cdef void write_value(self, Value value) except *:
        value.apply(_ValueWriter(self))

    cdef void write_bool(self, bint value) except *: pass
    cdef void write_byte(self, int8_t value) except *: pass
    cdef void write_double(self, double value) except *: pass
    cdef void write_i16(self, int16_t value) except *: pass
    cdef void write_i32(self, int32_t value) except *: pass
    cdef void write_i64(self, int64_t value) except *: pass
    cdef void write_binary(self, char* value, int32_t length) except *: pass
    cdef void write_struct_begin(self) except *: pass
    cdef void write_field_begin(self, FieldHeader header) except *: pass
    cdef void write_field_end(self) except *: pass
    cdef void write_struct_end(self) except *: pass
    cdef void write_map_begin(self, MapHeader header) except *: pass
    cdef void write_map_end(self) except *: pass
    cdef void write_set_begin(self, SetHeader header) except *: pass
    cdef void write_set_end(self) except *: pass
    cdef void write_list_begin(self, ListHeader header) except *: pass
    cdef void write_list_end(self) except *: pass
    cdef void write_message_begin(self, MessageHeader header) except *: pass
    cdef void write_message_end(self) except *: pass


cdef class ProtocolReader:

    # Skip

    cdef void skip(self, int typ) except *: pass
    cdef void skip_binary(self) except *: pass
    cdef void skip_map(self) except *: pass
    cdef void skip_list(self) except *: pass
    cdef void skip_set(self) except *: pass
    cdef void skip_struct(self) except *: pass

    # Primitives

    cdef bint read_bool(self) except *: pass
    cdef int8_t read_byte(self) except *: pass
    cdef double read_double(self) except *: pass
    cdef int16_t read_i16(self) except *: pass
    cdef int32_t read_i32(self) except *: pass
    cdef int64_t read_i64(self) except *: pass
    cdef bytes read_binary(self): pass

    # Structs

    cdef void read_struct_begin(self) except *: pass
    cdef FieldHeader read_field_begin(self) except *:
        """Parse the next three bytes as a FieldHeader object.

        :return: FieldHeader with type and id set to -1 if a struct end is found.
        """
        pass
    cdef void read_field_end(self) except *: pass
    cdef void read_struct_end(self) except *: pass

    # Containers

    cdef MapHeader read_map_begin(self) except *: pass
    cdef void read_map_end(self) except *: pass

    cdef SetHeader read_set_begin(self) except *: pass
    cdef void read_set_end(self) except *: pass

    cdef ListHeader read_list_begin(self) except *: pass
    cdef void read_list_end(self) except *: pass

    # Messages

    cdef MessageHeader read_message_begin(self): pass
    cdef void read_message_end(self) except *: pass


cdef class Protocol(object):
    """Base class for all protocol implementations.

    .. versionchanged:: 1.0

        Removed ``dumps`` and ``loads`` methods and added
        ``serialize_message`` and ``deserialize_message``.
    """

    cpdef ProtocolReader reader(self, ReadBuffer buff):
        raise NotImplementedError

    cpdef ProtocolWriter writer(self, WriteBuffer buff):
        raise NotImplementedError

    cpdef bytes serialize_value(self, Value value):
        """Serialize the given ``Value``.

        :param ~thriftrw.wire.Value value:
            Value to serialize.
        :returns:
            Serialized value.
        """
        cdef WriteBuffer buff = WriteBuffer()
        cdef ProtocolWriter writer = self.writer(buff)
        writer.write_value(value)
        return buff.value

    cpdef Value deserialize_value(self, int typ, bytes s):
        """Parse a ``Value`` of the given type.

        :param typ:
            Type code of the value to parse.
        :param s:
            Bytes to decode.
        :returns:
            Parsed :py:class:`~thriftrw.wire.Value`.
        :raises thriftrw.errors.ThriftProtocolError:
            If the object failed to deserialize.
        """

    cpdef bytes serialize_message(self, Message message):
        """Serialize a ``Message``.

        The message body must be a :py:class:`~thriftrw.wire.Value`.

        :param message:
            Message to serialize.
        :returns:
            Serialized message.
        """
        cdef WriteBuffer buff = WriteBuffer()
        cdef ProtocolWriter writer = self.writer(buff)
        cdef MessageHeader header = MessageHeader(
            message.name,
            message.message_type,
            message.seqid,
        )

        writer.write_message_begin(header)
        writer.write_value(message.body)
        writer.write_message_end()

        return buff.value

    cpdef Message deserialize_message(self, bytes s):
        """Deserialize a ``Message``.

        :param s:
            Bytes to decode
        :returns:
            Parsed :py:class:`~thriftrw.wire.Message` containing a
            :py:class:`~thriftrw.wire.Value` in its body.
        """

cdef class _ValueWriter(ValueVisitor):

    def __cinit__(_ValueWriter self, ProtocolWriter writer):
        self.writer = writer

    cdef object visit_bool(self, bint value):
        self.writer.write_bool(value)

    cdef object visit_byte(self, int8_t value):
        self.writer.write_byte(value)

    cdef object visit_double(self, double value):
        self.writer.write_double(value)

    cdef object visit_i16(self, int16_t value):
        self.writer.write_i16(value)

    cdef object visit_i32(self, int32_t value):
        self.writer.write_i32(value)

    cdef object visit_i64(self, int64_t value):
        self.writer.write_i64(value)

    cdef object visit_binary(self, bytes value):
        self.writer.write_binary(value, len(value))

    cdef object visit_struct(self, list fields):
        self.writer.write_struct_begin()
        for f in fields:
            self.writer.write_field_begin(FieldHeader(f.ttype, f.id))
            f.value.apply(self)
            self.writer.write_field_end()
        self.writer.write_struct_end()

    cdef object visit_map(self, int8_t key_ttype, int8_t value_ttype,
                          list pairs):
        cdef MapHeader header = MapHeader(key_ttype, value_ttype, len(pairs))
        self.writer.write_map_begin(header)
        for item in pairs:
            item.key.apply(self)
            item.value.apply(self)
        self.writer.write_map_end()

    cdef object visit_set(self, int8_t value_ttype, list values):
        cdef SetHeader header = SetHeader(value_ttype, len(values))
        self.writer.write_set_begin(header)
        for value in values:
            value.apply(self)
        self.writer.write_set_end()

    cdef object visit_list(self, int8_t value_ttype, list values):
        cdef ListHeader header = ListHeader(value_ttype, len(values))
        self.writer.write_list_begin(header)
        for value in values:
            value.apply(self)
        self.writer.write_list_end()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from collections import defaultdict

from thriftrw.wire cimport ttype
from thriftrw._cython cimport richcompare
from thriftrw.wire.value cimport I32Value, Value
from thriftrw.protocol.core cimport ProtocolWriter, ProtocolReader
from .base cimport TypeSpec
from . cimport check

from ..errors import ThriftCompilerError

__all__ = ['EnumTypeSpec']


cdef class EnumTypeSpec(TypeSpec):
    """TypeSpec for enum types.

    .. py:attribute:: name

        Name of the enum class.

    .. py:attribute:: items

        Mapping of enum item names to item values.

    .. py:attribute:: values_to_names

        Mapping of enum item values to a list of enum item names with that
        value.

        .. versionchanged:: 1.1

            Changed to a list of names.

    .. py:attribute:: surface

        The surface for this spec.

    The `surface` for enum types is a class with the following:

    .. py:attribute:: type_spec

        :py:class:`EnumTypeSpec` for the type.

    .. py:attribute:: items

        A tuple of the names of all items in this enum.

    .. py:attribute:: values

        A tuple of the values of all items in this enum in the same order as
        ``items``.

    .. py:classmethod:: name_of(value)

        Finds the name of an enum item by its value. If multiple enum items
        have this value, any of the matching item names could can be returned.

        :param int value:
            A value for an item defined in this enum.
        :returns:
            Name of the item with that value or None if no such item exists.

    And one attribute for each enum item that has the same name as the
    attribute and points to the value for that item.

    Given the definition,::

        enum Role {
            User, Admin
        }

    The generated class is roughly equivalent to,

    .. code-block:: python

        class Role(object):
            User = 0
            Admin = 1

            items = ('User', 'Admin')
            values = (0, 1)

            type_spec = # ...

            def name_of(self, value):
                # ...

    .. versionchanged:: 1.1

        Added support for multiple enum items with the same value.
    """

    def __init__(self, name, items):
        assert name
        assert items is not None

        self.ttype_code = ttype.I32
        self.name = str(name)
        self.items = items

        values_to_names = defaultdict(lambda: [])
        for name, value in items.items():
            values_to_names[value].append(name)

        self.values_to_names = values_to_names
        self.linked = False
        self.surface = None

    cpdef TypeSpec link(self, scope):
        if not self.linked:
            self.linked = True
            self.surface = enum_cls(self, scope)
        return self

    cpdef object read_from(EnumTypeSpec self, ProtocolReader reader):
        return reader.read_i32()

    cpdef Value to_wire(self, object value):
        return I32Value(value)

    cpdef object to_primitive(self, object value):
        return value

    cpdef object from_wire(self, Value wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value

    cpdef void write_to(EnumTypeSpec self, ProtocolWriter writer,
                        object value) except *:
        writer.write_i32(value)

    cpdef object from_primitive(self, object prim_value):
        val = self.items.get(prim_value)
        if val is not None:
            return val
        return prim_value

    cpdef void validate(self, object instance) except *:
        if instance not in self.values_to_names:
            raise ValueError(
                '%r is not a valid value for enum "%s"' % (instance, self.name)
            )

    @classmethod
    def compile(cls, enum):
        items = {}

        # TODO check for int32 overflow.

        prev = -1
        for item in enum.items:
            value = item.value
            if value is None:
                value = prev + 1
            prev = value

            if item.name in items:
                raise ThriftCompilerError(
                    'Enum entry "%s.%s" on line %d has duplicates.' %
                    (enum.name, item.name, item.lineno)
                )

            items[item.name] = value

        return cls(enum.name, items)

    def __richcmp__(EnumTypeSpec self, EnumTypeSpec other not None, int op):
        return richcompare(op, [
            (self.name, other.name),
            (self.items, other.items),
        ])

    def __str__(self):
        return 'EnumTypeSpec(name=%r, items=%r)' % (self.name, self.items)

    def __repr__(self):
        return str(self)


def enum_cls(enum_spec, scope):
    """Generates a class for the given EnumTypeSpec.

    :param EnumTypeSpec enum_spec:
        EnumTypeSpec for which the class is being generated.
    """
    enum_dct = {}

    for name, item in enum_spec.items.items():
        enum_dct[name] = item

    keys, values = zip(*enum_spec.items.items())

    enum_dct['_VALUES_TO_NAMES'] = enum_spec.values_to_names
    enum_dct['items'] = tuple(keys)
    enum_dct['values'] = tuple(values)
    enum_dct['name_of'] = name_of
    enum_dct['type_spec'] = enum_spec
    enum_dct['__thrift_module__'] = scope.module
    enum_dct['__slots__'] = ()
    enum_dct['__doc__'] = enum_docstring(
        enum_spec.name, enum_spec.items.items()
    )

    return type(str(enum_spec.name), (object,), enum_dct)


@classmethod
def name_of(cls, value):
    """Returns the name of the enum item with the given value.

    :param int value:
        Enum item value
    :returns:
        Name of the corresponding enum item or None if no such item exists.
    """
    names = cls._VALUES_TO_NAMES.get(value)
    if names:
        return names[0]
    return None


def enum_docstring(name, items):
    item_section = '\n'.join(
        '-    %s = %s' % pair for pair in items
    )
    return '%s defines the following items:\n\n' % name + item_section



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from cpython.mem cimport (
    PyMem_Malloc,
    PyMem_Realloc,
    PyMem_Free,
)
from libc.string cimport memcpy

from .errors import EndOfInputError

cdef int DEFAULT_CAPACITY = 4096  # 4k


cdef class WriteBuffer(object):
    """A write-only in-memory buffer.

    It defaults to a 4k sized buffer.

    The ``value`` attribute makes all data written to the buffer so far
    available as a Python ``bytes`` object.
    """

    def __cinit__(self, int init_capacity=0):
        """
        :param int init_capacity:
            Initial capacity of the write buffer.
        """
        init_capacity = init_capacity or DEFAULT_CAPACITY

        self.data = <char*>PyMem_Malloc(init_capacity)
        self.length = 0
        self.capacity = init_capacity

    def __dealloc__(self):
        if self.data != NULL:
            PyMem_Free(self.data)
            self.data = NULL

    cpdef void clear(self):
        """Clears the buffer."""

        capacity = self.length + self.capacity
        self.length = 0
        self.capacity = capacity

    cpdef void write_bytes(self, bytes data):
        """Writes the given Python bytes object to the buffer.

        :param bytes data:
            Data to write to the buffer.
        """
        self.write(data, len(data))

    cdef void write(self, char* data, int count):
        """Writes bytes from the given memory block to the buffer.

        :param int data:
            Pointer to source data block.
        :param int count:
            Number of bytes to write.
        """
        self.ensure_capacity(count)

        memcpy(self.data + self.length, data, count)
        self.length += count
        self.capacity -= count

    cdef void ensure_capacity(self, int min_bytes):
        """Ensures that the buffer has enough room for at least ``min_bytes``
        more bytes.
        """
        if min_bytes <= self.capacity:
            return

        cdef int new_total_length = self.length * 2
        if new_total_length - self.length < min_bytes:
            # If adding as much room as the current buffer size is not enough,
            # add just enough room on top.
            new_total_length += min_bytes

        self.data = <char*>PyMem_Realloc(self.data, new_total_length)
        self.capacity = new_total_length - self.length

    property value:
        """Data written to the buffer so far."""

        def __get__(self):
            cdef bytes out = self.data[:self.length]
            return out


cdef class ReadBuffer(object):
    """A read-only in-memory buffer."""

    def __cinit__(self, bytes data):
        """Initialize a ReadBuffer that reads from the given ``bytes``.

        :param bytes data:
            Block of data that this read buffer will yield.
        """
        self._data = data
        self.data = data

        self.offset = 0
        self.length = len(data)

    cpdef void read(self, char* dest, int count) except *:
        """Reads ``count`` bytes into ``dest``.

        :param dest:
            Memory block to which the bytes will be written.
        :param int count:
            Number of bytes to read.
        :raises EndOfInputError:
            If the number of bytes available in the buffer is less than
            ``count``.
        """
        if count > self.length - self.offset:
            raise EndOfInputError(
                'Expected %d bytes but got %d bytes.' %
                (count, self.available)
            )

        memcpy(dest, self.data + self.offset, count)
        self.offset += count

    cpdef bytes take(self, int count):
        """Read the next ``count`` bytes from the buffer.

        :param int count:
            Number of bytes to read.
        :returns:
            ``bytes`` object containing exactly ``count`` bytes.
        :raises EndOfInputError:
            If the number of bytes available in the buffer is less than
            ``count``.
        """
        if count > self.length - self.offset:
            raise EndOfInputError(
                'Expected %d bytes but got %d bytes.' %
                (count, self.available)
            )

        cdef bytes result = self.data[self.offset:self.offset + count]
        self.offset += count

        return result

    cpdef void skip(self, int count) except *:
        """Skip ``count`` bytes without performing any memory copying.

        :param int count:
            Number of bytes to read.
        :raises EndOfInputError:
            If the number of bytes available in the buffer is less than
            ``count``.
        """
        if count > self.length - self.offset:
            raise EndOfInputError(
                'Expected %d bytes but got %d bytes.' %
                (count, self.available)
            )

        self.offset += count

    property available:
        """Number of bytes available in the buffer."""

        def __get__(self):
            return self.length - self.offset



from __future__ import absolute_import, unicode_literals, print_function

import time
import os.path

from tornado import web
from tornado.ioloop import IOLoop
from tornado.httpserver import HTTPServer

import thriftrw

ping = thriftrw.load(
    os.path.join(os.path.dirname(__file__), 'ping.thrift'),
)


class ThriftRequestHandler(web.RequestHandler):

    def post(self):
        assert self.request.body

        message = ping.loads.message(ping.Ping, self.request.body)
        method, handler = self._METHODS[message.name]

        args = message.body
        resp = method.response(success=handler(self, args))

        reply = ping.dumps.message(resp, seqid=message.seqid)
        self.write(reply)

    def handle_ping(self, args):
        print('Hello, %s' % args.name)
        return ping.Pong(time.time())

    _METHODS = {'ping': (ping.Ping.ping, handle_ping)}


if __name__ == "__main__":
    app = web.Application([
        (r'/thrift', ThriftRequestHandler),
    ])
    HTTPServer(app).listen(8888)
    print('Listening on http://127.0.0.1:8888/thrift')
    IOLoop.current().start()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from thriftrw.wire.value cimport Value
from thriftrw.protocol.core cimport ProtocolWriter, ProtocolReader

__all__ = ['TypeSpec']


cdef class TypeSpec:
    """Base class for classes representing TypeSpecs.

    A TypeSpec knows how to convert values of the corresponding type to and
    from :py:class:`thriftrw.wire.Value` objects.
    """
    __slots__ = ()

    property name:
        """Name of the type referenced by this type spec."""

        def __get__(self):
            raise NotImplementedError

    property surface:
        """The surface of a type spec.

        The surface of a type spec, if non-None is the value associated with
        its name at the top-level in the generated module.
        """

        def __get__(self):
            raise NotImplementedError

    cpdef object read_from(TypeSpec self, ProtocolReader reader):
        """Read a primitive value of this type from :py:class:`thriftrw.protocol.ProtocolReader`."""
        raise NotImplementedError

    cpdef void write_to(TypeSpec self, ProtocolWriter writer,
                        object value) except *:
        """Writes a value directly to :py:class:`thriftrw.protocol.ProtocolWriter`."""
        writer.write_value(self.to_wire(value))

    cpdef Value to_wire(TypeSpec self, object value):
        """Converts the given value into a :py:class:`thriftrw.wire.Value`
        object.

        :returns thriftrw.wire.Value:
            Wire representation of the value.
        """
        raise NotImplementedError(
            'to_wire called on unlinked type reference: %r', self
        )

    cpdef object from_wire(TypeSpec self, Value wire_value):
        """Converts the given :py:class:`thriftrw.wire.Value` back into the
        original type.

        :param thriftr.wire.Value wire_value:
            Value to convert.
        :raises ValueError:
            If the type of the wire value does not have the correct Thrift
            type for this type spec.
        """
        raise NotImplementedError(
            'from_wire called on unlinked type reference: %r', self
        )

    cpdef TypeSpec link(self, scope):
        raise NotImplementedError

    cpdef object to_primitive(TypeSpec self, object value):
        """Converts a value matching this type spec into a primitive value.

        A primitive value is a text, binary, integer, or float value, or a
        list or dict of other primitive values.

        .. versionadded:: 0.4

        :param value:
            Value matching this TypeSpec.
        :returns:
            A representation of that value using only primitive types, lists,
            and maps.
        """
        raise NotImplementedError(
            'to_primitive called on unlinked type reference: %r', self
        )

    cpdef object from_primitive(TypeSpec self, object prim_value):
        """Converts a primitive value into a value of this type.

        A primitive value is a text, binary, integer, or float value, or a
        list or dict of other primitive values.

        .. versionadded:: 0.4

        :param prim_value:
            A primitive value as produced by ``to_primitive``.
        :returns:
            A value matching this TypeSpec.
        """
        raise NotImplementedError(
            'from_primitive called on unlinked type reference: %r', self
        )

    cpdef void validate(TypeSpec self, object o) except *:
        """Whether an instance of this spec is valid.

        :param instance:
            An instance of the type described by this spec.

        :raises TypeError:
            If the value did not match the type expected by this TypeSpec.
        :raises ValueError:
            If the value matched the type but did not hold the correct set of
            acceptable values.
        """
        raise NotImplementedError(
            'validate called on unlinked type reference: %r', self
        )



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import copy

from thriftrw.wire cimport ttype
from thriftrw.wire.value cimport Value
from thriftrw._cython cimport richcompare
from thriftrw.wire.value cimport StructValue
from thriftrw.protocol.core cimport (
    ProtocolWriter,
    FieldHeader,
    ProtocolReader,
)
from thriftrw.spec.union cimport UnionTypeSpec
from .base cimport TypeSpec
from .field cimport FieldSpec
from . cimport check

from . import common
from ..errors import ThriftCompilerError

__all__ = ['StructTypeSpec', 'FieldSpec']


cdef class StructTypeSpec(TypeSpec):
    """A struct is a collection of named fields.

    .. py:attribute:: name

        Name of the struct.

    .. py:attribute:: fields

        Collection of :py:class:`FieldSpec` objects representing the fields of
        this struct.

    The surface for struct types is a class with the following:

    .. py:attribute:: type_spec

        :py:class:`StructTypeSpec` for the type.

    .. py:method:: __init__(self, *args, **kwargs)

        Accepts all fields of the struct as arguments. Required arguments are
        placed first and may be specified as positional arguments (in the same
        order as they appear in the IDL). Optional arguments and required
        arguments come next. A TypeError will be raised if the required
        arguments for the struct are not filled with non-None values.

    .. py:method:: to_primitive(self)

        Converts the struct into a dictionary mapping field names to primitive
        representation of field values.

        Only the following types are used in primitive representations:
        ``bool``, ``bytes``, ``float``, ``str`` (``unicode`` in Python < 3),
        ``int``, ``long``, ``dict``, ``list``.

        .. versionadded:: 0.4

    .. py:classmethod:: from_primitive(cls, value)

        Converts a dictionary holding a primitive representation of a value of
        this type (as returned by ``to_primitive``) into an instance of this
        class.

        .. versionadded:: 0.4

    And obvious definitions of ``__str__`` and ``__eq__``.

    Given the definition,::

        struct User {
            1: required string name
            2: optional string email
            3: required bool isActive = true
        }

    A class roughly equivalent to the following is generated,

    .. code-block:: python

        class User(object):

            __slots__ = ('name', 'email', 'isActive')

            type_spec = # ...

            def __init__(name, email=None, isActive=True):
                # ...
    """

    def __init__(self, name, fields, base_cls=None):
        """Initialize a new StructTypeSpec.

        :param str name:
            Name of the Struct.
        :param fields:
            Collection of :py:class:`FieldSpec` objects.
        :param base_cls:
            Base class to use for generates classes. Defaults to ``object``.
        """

        self.ttype_code = ttype.STRUCT
        self.name = str(name)
        self.fields = fields
        self.linked = False
        self.surface = None
        self.base_cls = base_cls or object
        self._index = {}

    cpdef TypeSpec link(self, scope):
        if not self.linked:
            self.linked = True
            self.fields = [field.link(scope) for field in self.fields]
            self.surface = struct_cls(self, scope)
            for field in self.fields:
                self._index[field.id] = field

        return self

    @classmethod
    def compile(cls, struct, require_requiredness=True):
        fields = []
        ids = set()
        names = set()
        for field in struct.fields:
            if field.name in names:
                raise ThriftCompilerError(
                    'Field "%s" of struct "%s" on line %d has duplicates.' %
                    (field.name, struct.name, field.lineno)
                )
            names.add(field.name)

            if field.id in ids:
                raise ThriftCompilerError(
                    'Field ID "%d" of struct "%s" on line %d '
                    'has already been used.' % (
                        field.id, struct.name, field.lineno
                    )
                )
            ids.add(field.id)
            fields.append(FieldSpec.compile(
                field=field,
                struct_name=struct.name,
                require_requiredness=require_requiredness,
            ))
        return cls(struct.name, fields)

    cpdef object read_from(StructTypeSpec self, ProtocolReader reader):
        reader.read_struct_begin()

        cdef dict kwargs = {}
        cdef object val
        cdef FieldSpec field
        cdef FieldHeader header = reader.read_field_begin()

        while header.type != -1:
            # You'd be surprised how much faster the CallOneArg path is compared to
            # the unoptimized version. Make sure this never reades .get(val, None).
            field = self._index.get(header.id)

            # Unrecognized field--possibly different version of struct definition.
            if field is None or field.spec.ttype_code != header.type:
                reader.skip(header.type)
            else:
                kwargs[field.name] = field.spec.read_from(reader)

            reader.read_field_end()
            header = reader.read_field_begin()

        reader.read_struct_end()
        return self.surface(**kwargs)

    cpdef Value to_wire(self, object struct):
        fields = []

        for field in self.fields:
            value = getattr(struct, field.name)
            if value is None:
                continue
            fields.append(field.to_wire(value))

        return StructValue(fields)

    cpdef object to_primitive(self, object struct):
        prim = {}

        for field in self.fields:
            value = getattr(struct, field.name)
            if value is None:
                continue

            prim[field.name] = field.spec.to_primitive(value)
        return prim

    cpdef void write_to(StructTypeSpec self, ProtocolWriter writer,
                        object struct) except *:
        # This is the single most performance critical method in the serialization path.
        #
        # Carefully inspect benchmark results if you make changes here and avoid python
        # API interaction at any cost.
        writer.write_struct_begin()
        cdef FieldHeader header

        for field in self.fields:
            value = getattr(struct, field.name)
            if value is None:
                continue

            header.type = (<FieldSpec> field).spec.ttype_code
            header.id = (<FieldSpec> field).id

            writer.write_field_begin(header)
            (<FieldSpec> field).spec.write_to(writer, value)
            writer.write_field_end()

        writer.write_struct_end()

    cpdef object from_wire(self, Value wire_value):
        check.type_code_matches(self, wire_value)
        kwargs = {}
        for field in self.fields:
            field_value = wire_value.get(field.id, field.ttype_code)
            if field_value is None:
                continue
            kwargs[field.name] = field.from_wire(field_value)

        return self.surface(**kwargs)

    cpdef object from_primitive(self, object prim_value):
        kwargs = {}

        for field in self.fields:
            field_value = prim_value.get(field.name)
            if field_value is None:
                continue
            kwargs[field.name] = field.spec.from_primitive(field_value)

        return self.surface(**kwargs)

    cpdef void validate(self, object instance) except *:
        check.instanceof_surface(self, instance)

        for field in self.fields:
            field_value = getattr(instance, field.name)
            if field_value is None:
                if field.required:
                    raise TypeError(
                        'Field "%s" of "%s" is required. It cannot be None.'
                        % (field.name, self.name)
                    )
                else:
                    continue

            # Since we validate at construction time, child structs are
            # almost certainly valid unless consumers are directly mutating
            # thrift structs. As an optimization, avoid recursively revalidating
            # these.
            if field.spec.ttype_code == ttype.STRUCT:
                check.instanceof_surface(field.spec, field_value)
                continue

            try:
                field.spec.validate(field_value)
            except (ValueError, TypeError) as e:
                raise e.__class__(
                    'Field %d of %s is invalid: %s' % (field.id, self.name, e)
                )

    def __str__(self):
        return '%s(name=%r, fields=%r)' % (
            self.__class__.__name__, self.name, self.fields
        )

    def __repr__(self):
        return str(self)

    def __richcmp__(StructTypeSpec self, StructTypeSpec other not None, int op):
        return richcompare(op, [
            (self.name, other.name),
            (self.fields, other.fields),
            (self.base_cls, other.base_cls),
        ])


def struct_init(cls_name, field_names, field_defaults, base_cls, fields):
    """Generate the ``__init__`` method for structs.

    ``field_names`` is a list or tuple of field names for the constructor
    in-order. ``field_defaults`` is a list or tuple of default values for the
    last ``len(field_defaults)`` fields in-order.

    That is, for a signature like, ``f(a, b, c=1, d=2)``, ``field_names`` is
    ``('a', 'b', 'c', 'd')`` and ``field_defaults`` is ``(1, 2)``.

    :param cls_name:
        Name of the class. This is used for error messages.
    :param field_names:
        List or tuple of field names of the struct.
    :param field_defaults:
        List or tuple of default values for the last ``len(field_defaults)``
        fields. For optional fields with no default value, ``field_defaults``
        should have a ``None`` default.
    :param fields:
        Map of field name to ``FieldSpec`` objects describing this struct.
    :returns:
        Implementation of the ``__init__`` method for that struct.
    """
    field_defaults = field_defaults or ()

    # Number of fields on the struct. We can't accept more than this many
    # arguments.
    cdef int num_fields = len(field_names)

    # Number of fields with default values.
    cdef int num_defaults = len(field_defaults)

    assert num_defaults <= num_fields, 'Too many defaults provided.'

    # The last num_defaults fields have default values, so all other fields
    # must have arguments passed in.
    cdef int num_required = num_fields - num_defaults

    def __init__(self, *args, **kwargs):
        base_cls.__init__(self)

        cdef int num_args = len(args)
        cdef int num_kwargs = len(kwargs)
        cdef int num_total = num_args + num_kwargs

        # This argument binding implementation was adapted from
        # inspect.getcallargs() with adjustments for our requirements of
        # non-None fields.

        # TODO pull out argument binding logic into separate function
        # elsewhere.

        if num_total < num_required:
            raise TypeError(
                '%s() takes at least %d arguments (%d given)' % (
                    cls_name, num_required + 1, num_total + 1
                    # + 1 for 'self'
                )
            )

        if num_total > num_fields:
            raise TypeError(
                '%s() takes at most %d arguments (%d given)' % (
                    cls_name, num_fields + 1, num_total + 1
                    # + 1 for 'self'
                )
            )

        cdef object value
        cdef FieldSpec field_spec
        cdef int i = -1
        for name in field_names:
            i += 1
            # Check positionals
            if i < num_args:
                value = args[i]

                # If found in positionals, it can't be in kwargs
                if name in kwargs:
                    raise TypeError(
                        '%s() got multiple values for argument "%s"'
                        % (cls_name, name)
                    )
            else:
                value = kwargs.pop(name, None)

            # We do validation inline for structs as an optimization
            field_spec = <FieldSpec> fields[name]
            if value is None:
                if field_spec.default_value is not None:
                    value = copy.deepcopy(field_spec.default_value)
                elif field_spec.required:
                    raise TypeError(
                        'Field "%s" of "%s" is required. It cannot be None.'
                        % (name, self.__class__.__name__)
                    )

            # Since we validate at construction time, child structs are
            # almost certainly valid unless consumers are directly mutating
            # thrift structs. As an optimization, avoid recursively revalidating
            # these.
            elif field_spec.spec.ttype_code == ttype.STRUCT:
                # TODO: Avoid instance of, just do a type() is ... check.
                check.instanceof_surface(field_spec.spec, value)
            else:
                try:
                    field_spec.spec.validate(value)
                except (ValueError, TypeError) as e:
                    raise e.__class__(
                        'Field %s of %s is invalid: %s' % (
                            field_spec.name,
                            self.__class__.__name__,
                            e,
                        )
                    )

            setattr(self, name, value)

        # TODO: Is len() == 0 faster?
        if kwargs:
            # Too many kwargs given.
            for name in kwargs:
                raise TypeError(
                    '%s() got an unexpected keyword argument "%s"'
                    % (cls_name, name)
                )

    return __init__


def struct_docstring(cls_name, required_fields, optional_fields):
    """Generates a docstring for generated structs."""
    required_section = ''
    if required_fields:
        required_section = '\n'.join(
            [''] +
            ['\nThe following arguments are required:\n'] +
            ['-   %s' % name for name in required_fields]
        )

    optional_section = ''
    if optional_fields:
        optional_section = '\n'.join(
            [''] +
            ['\nThe following arguments are optional:\n'] +
            ['-   %s (default: %s)' % pair for pair in optional_fields]
        )

    param_list = (
        list(required_fields) +
        ['%s=None' % name for name, value in optional_fields]
    )

    header = '%s(%s)' % (cls_name, ', '.join(param_list))
    return header + required_section + optional_section


def struct_cls(struct_spec, scope):
    """Generate a class for a struct.

    The generated class has the following attributes:

    :param StructTypeSpec struct_spec:
        Type specification of the struct.
    :param thriftrw.compile.Scope scope:
        The compilation scope
    """
    slots = set()

    required_fields = []
    optional_fields = []
    field_defaults = []

    for field in struct_spec.fields:
        slots.add(field.name)

        default = None
        if field.default_value is not None:
            default = field.default_value

        if field.required and default is None:
            # required fields that have default values are optional as far as
            # the constructor is concerned.
            required_fields.append(field.name)
        else:
            optional_fields.append(field.name)
            field_defaults.append(default)

    field_names = required_fields + optional_fields
    field_map = {spec.name: spec for spec in struct_spec.fields}

    struct_dct = {}
    struct_dct['type_spec'] = struct_spec
    struct_dct['to_primitive'] = common.to_primitive_method(struct_spec)
    struct_dct['from_primitive'] = common.from_primitive_classmethod()
    struct_dct['__thrift_module__'] = scope.module
    struct_dct['__slots__'] = tuple(slots)
    struct_dct['__init__'] = struct_init(
        struct_spec.name,
        field_names,
        field_defaults,
        struct_spec.base_cls,
        field_map,
    )
    struct_dct['__str__'] = common.fields_str(struct_spec.name, field_names)
    struct_dct['__repr__'] = struct_dct['__str__']
    struct_dct['__eq__'] = common.fields_eq(set(field_names))
    struct_dct['__doc__'] = struct_docstring(
        struct_spec.name,
        required_fields,
        list(zip(optional_fields, field_defaults)),
    )

    # TODO generate a reasonable docstring for the class too.

    return type(str(struct_spec.name), (struct_spec.base_cls,), struct_dct)



from __future__ import absolute_import

from thrift.transport.THttpClient import THttpClient
from thrift.protocol.TBinaryProtocol import TBinaryProtocolAccelerated

from ping import Ping


trans = THttpClient('http://localhost:8888/thrift')
proto = TBinaryProtocolAccelerated(trans)
client = Ping.Client(proto)

print client.ping('world')



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from thriftrw._cython cimport richcompare
from thriftrw.wire.value cimport Value

from .base cimport TypeSpec

__all__ = ['TypeReference']


cdef class TypeReference(TypeSpec):
    """A reference to another type."""

    __slots__ = ('name', 'lineno')

    def __init__(self, name, lineno):
        self.name = str(name)
        self.lineno = lineno

    cpdef TypeSpec link(self, scope):
        return scope.resolve_type_spec(self.name, self.lineno)

    # It may be worth making this implement the TypeSpec interface and raise
    # exceptions complaining about unresolved type references, since that's
    # probably a bug.

    def __str__(self):
        return 'TypeReference(%s, lineno=%d)' % (
            self.name, self.lineno
        )

    def __repr__(self):
        return str(self)

    def __richcmp__(TypeReference self, TypeReference other not None, int op):
        return richcompare(op, [
            (self.name, other.name),
            (self.lineno, other.lineno),
        ])



from __future__ import absolute_import, unicode_literals, print_function

import os.path
import requests

import thriftrw

ping = thriftrw.load(
    os.path.join(os.path.dirname(__file__), 'ping.thrift'),
)


def main():
    req = ping.Ping.ping.request('world')

    response = requests.post(
        'http://127.0.0.1:8888/thrift',
        data=ping.dumps.message(req, seqid=42),
    )
    reply = ping.loads.message(ping.Ping, response.content)
    assert reply.name == 'ping'
    assert reply.seqid == 42

    resp = reply.body
    print(resp)

if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os
import re

from setuptools import setup
from setuptools import find_packages
from setuptools.command.sdist import sdist as _sdist
from setuptools.extension import Extension as _Extension

cmdclass = {}
ext_modules = []

cython_modules = [
    'thriftrw._buffer',
    'thriftrw._cython',
    'thriftrw._runtime',
    'thriftrw.protocol.core',
    'thriftrw.protocol.binary',
    'thriftrw.spec.base',
    'thriftrw.spec.check',
    'thriftrw.spec.common',
    'thriftrw.spec.enum',
    'thriftrw.spec.exc',
    'thriftrw.spec.field',
    'thriftrw.spec.list',
    'thriftrw.spec.map',
    'thriftrw.spec.reference',
    'thriftrw.spec.primitive',
    'thriftrw.spec.service',
    'thriftrw.spec.set',
    'thriftrw.spec.spec_mapper',
    'thriftrw.spec.struct',
    'thriftrw.spec.typedef',
    'thriftrw.spec.union',
    'thriftrw.wire.message',
    'thriftrw.wire.mtype',
    'thriftrw.wire.ttype',
    'thriftrw.wire.value',
]

extension_extras = {}

# If Cython is available we will re-cythonize the pyx files, otherwise we just
# compile the packaged C files.
extension_filetype = '.c'

Extension = None
try:
    import Cython.Distutils

    # from __future__ import absolute_import needs cython >= 0.17
    if tuple(int(v) for v in Cython.__version__.split('.')) >= (0, 17, 0):
        cmdclass.update(build_ext=Cython.Distutils.build_ext)

        # Check if we forgot to add something to cython_modules.
        for root, _, files in os.walk('thriftrw'):
            for name in files:
                if not name.endswith('.pyx'):
                    continue
                path = os.path.join(root, name)
                module = path.replace('/', '.')[:-4]
                if module not in cython_modules:
                    raise Exception(
                        'Module "%s" (%s) is not present in the '
                        '"cython_modules" list.'
                        % (module, path)
                    )

        Extension = Cython.Distutils.Extension
        extension_filetype = '.pyx'

        cython_directives = {
            'embedsignature': True,
        }

        if os.getenv('THRIFTRW_PROFILE'):
            # Add hooks for the profiler in the generated C code.
            cython_directives['profile'] = True

        if os.getenv('THRIFTRW_COVERAGE'):
            # Add line tracing hooks to the generated C code. The hooks aren't
            # actually enabled unless the CYTHON_TRACE macre is also set. This
            # affects performance negatively and should only be used during
            # testing.
            extension_extras['define_macros'] = [('CYTHON_TRACE', '1')]
            cython_directives['linetrace'] = True

        if cython_directives:
            extension_extras['cython_directives'] = cython_directives
except ImportError:
    pass

if Extension is None:
    Extension = _Extension


for module in cython_modules:
    ext_modules.append(
        Extension(
            module,
            [module.replace('.', '/') + extension_filetype],
            **extension_extras
        )
    )


class sdist(_sdist):
    """This forces us to always re-compile extensions before releasing."""

    def run(self):
        try:
            from Cython.Build import cythonize

            cythonize([
                module.replace('.', '/') + '.pyx' for module in cython_modules
            ])
        except ImportError:
            pass
        _sdist.run(self)

cmdclass['sdist'] = sdist


version = None
with open('thriftrw/__init__.py', 'r') as f:
    for line in f:
        m = re.match(r'^__version__\s*=\s*(["\'])([^"\']+)\1', line)
        if m:
            version = m.group(2)
            break

if not version:
    raise Exception(
        'Could not determine version number from thriftrw/__init__.py'
    )


with open('README.rst') as f:
    long_description = f.read()

setup(
    name='thriftrw',
    version=version,
    description=(
        'A library to serialize and deserialize Thrift values.'
    ),
    long_description=long_description,
    author='Abhinav Gupta',
    author_email='abg@uber.com',
    url='https://github.com/thriftrw/thriftrw-python',
    packages=find_packages(exclude=('tests', 'tests.*')),
    license='MIT',
    install_requires=['six', 'ply'],
    tests_require=['pytest', 'mock'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    ext_modules=ext_modules,
    cmdclass=cmdclass,
)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import sys
import inspect
import os.path

from .compile import Compiler
from .protocol import BinaryProtocol


class Loader(object):
    """Loads and compiles Thrift files."""

    __slots__ = ('compiler',)

    def __init__(self, protocol=None, strict=None, include_as=None):
        """Initialize a loader.

        :param thriftrw.protocol.Protocol protocol:
            The protocol to use to serialize and deserialize types. Defaults
            to the binary protocol.

        :param bool strict:
            When ``True`` (the default behavior) this enforces that "required"
            or "optional" always be specified for struct elements. This is
            desirable because Apache Thrift generates different behaviors for
            elements without "required" or "optional" depending on the
            language. Setting this to ``False`` is only recommended for
            compatibility with existing Thrift files.

        :param bool include_as:
            Whether thriftrw's custom include-as syntax is supported. Defaults
            to False. Note that this makes your Thrift files incomptabile with
            Apache Thrift. Use at your own risk.
        """
        protocol = protocol or BinaryProtocol()
        self.compiler = Compiler(
            protocol, strict=strict, include_as=include_as
        )

    def loads(self, name, document):
        """Parse and compile the given Thrift document.

        :param str name:
            Name of the Thrift document.
        :param str document:
            The Thrift IDL as a string.
        """
        return self.compiler.compile(name, document).link().surface

    def load(self, path, name=None):
        """Load and compile the given Thrift file.

        :param str path:
            Path to the ``.thrift`` file.
        :param str name:
            Name of the generated module. Defaults to the base name of the
            file.
        :returns:
            The compiled module.
        """
        if name is None:
            name = os.path.splitext(os.path.basename(path))[0]
            # TODO do we care if the file extension is .thrift?

        with open(path, 'r') as f:
            document = f.read()

        return self.compiler.compile(name, document, path).link().surface


_DEFAULT_LOADER = Loader(protocol=BinaryProtocol())

#: Parses and compiles the given Thrift file.
#:
#: Uses the binary protocol to serialize and deserialize values.
#:
#: For more advanced use, see :py:class:`thriftrw.loader.Loader`.
#:
#: :param str path:
#:     Path to the ``.thrift`` file.
#: :param str name:
#:     Name of the generated module. Defaults to the basename of the file.
#: :param bool force:
#:     Whether to ignore the cache and load the file anew. Defaults to False.
#: :raises thriftrw.errors.ThriftCompilerError:
#:     If there was an error compiling the Thrift file.
load = _DEFAULT_LOADER.load


def install(path, name=None):
    """Compiles a Thrift file and installs it as a submodule of the caller.

    Given a tree organized like so::

        foo/
            __init__.py
            bar.py
            my_service.thrift

    You would do,

    .. code-block:: python

        my_service = thriftrw.install('my_service.thrift')

    To install ``my_service`` as a submodule of the module from which you made
    the call. If the call was made in ``foo/bar.py``, the compiled
    Thrift file will be installed as ``foo.bar.my_service``. If the call was
    made in ``foo/__init__.py``, the compiled Thrift file will be installed as
    ``foo.my_service``. This allows other modules to import ``from`` the
    compiled module like so,

    .. code-block:: python

        from foo.my_service import MyService

    .. versionadded:: 0.2

    :param path:
        Path of the Thrift file. This may be an absolute path, or a path
        relative to the Python module making the call.
    :param str name:
        Name of the submodule. Defaults to the basename of the Thrift file.
    :returns:
        The compiled module
    """
    if name is None:
        name = os.path.splitext(os.path.basename(path))[0]

    callermod = inspect.getmodule(inspect.stack()[1][0])
    name = '%s.%s' % (callermod.__name__, name)

    if name in sys.modules:
        return sys.modules[name]

    if not os.path.isabs(path):
        callerfile = callermod.__file__
        path = os.path.normpath(
            os.path.join(os.path.dirname(callerfile), path)
        )

    sys.modules[name] = mod = load(path, name=name)
    return mod


__all__ = ['Loader', 'load', 'install']



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function


cdef bint richcompare_op(int op, int compare):
    # op    operation
    # 0     <
    # 1     <=
    # 2     ==
    # 3     !=
    # 4     >
    # 5     >=

    if op == 2:
        return compare == 0
    elif op == 3:
        return compare != 0
    elif op == 4:
        return compare > 0
    elif op == 5:
        return compare >= 0
    elif op == 0:
        return compare < 0
    elif op == 1:
        return compare <= 0
    else:
        assert False, 'Invalid comparison operator "%d"' % (op,)


cdef bint richcompare(int op, list pairs):
    """Utility function to make ``richcmp`` easier to write.

    It takes a list of attribute pairs. Attributes are compared in the order
    they appear and the result is returned based on the ``op``.

    .. code-block:: python

        def __richcmp__(self, other, op):
            return richcompare(
                op,
                [
                    (self.attr1, other.attr1),
                    (self.attr2, other.attr2),
                    # ...
                ]
            )
    """

    cdef int compare = 0

    for (left, right) in pairs:
        if left > right:
            compare = 1
            break
        elif left < right:
            compare = -1
            break

    return richcompare_op(op, compare)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Different TType codes supported by Thrift.

.. py:data:: BOOL
    :annotation: = 2

.. py:data:: BYTE
    :annotation: = 3

.. py:data:: DOUBLE
    :annotation: = 4

.. py:data:: I16
    :annotation: = 6

.. py:data:: I32
    :annotation: = 8

.. py:data:: I64
    :annotation: = 10

.. py:data:: BINARY
    :annotation: = 11

.. py:data:: STRUCT
    :annotation: = 12

.. py:data:: MAP
    :annotation: = 13

.. py:data:: SET
    :annotation: = 14

.. py:data:: LIST
    :annotation: = 15

"""

from __future__ import absolute_import, unicode_literals, print_function


cpdef str name_of(int value):
    """Returns the name of the TType with the given value.

    Returns None if no such TType exists.
    """
    if value == BOOL:
        return str('BOOL')
    elif value == BYTE:
        return str('BYTE')
    elif value == DOUBLE:
        return str('DOUBLE')
    elif value == I16:
        return str('I16')
    elif value == I32:
        return str('I32')
    elif value == I64:
        return str('I64')
    elif value == BINARY:
        return str('BINARY')
    elif value == STRUCT:
        return str('STRUCT')
    elif value == MAP:
        return str('MAP')
    elif value == SET:
        return str('SET')
    elif value == LIST:
        return str('LIST')
    else:
        return None
